"""Thor based player
======================

This player can play Thor cameras using :mod:`thorcam`.
"""

from time import perf_counter as clock
from queue import Queue, Empty
from os.path import splitext, join, exists, isdir, abspath, dirname

from kivy.clock import Clock
from kivy.properties import (
    NumericProperty, ReferenceListProperty,
    ObjectProperty, ListProperty, StringProperty, BooleanProperty,
    DictProperty, AliasProperty, OptionProperty, ConfigParserProperty)
from kivy.uix.boxlayout import BoxLayout
from kivy.logger import Logger
from kivy.lang import Builder

from cpl_media.player import BasePlayer, VideoMetadata
from cpl_media import error_guard
import cpl_media

try:
    from thorcam.camera import ThorCamClient
except ImportError as err:
    ThorCamClient = object
    Logger.debug('cpl_media: Could not import thorcam: {}'.format(err))


__all__ = ('ThorCamPlayer', 'ThorCamSettingsWidget')


class ThorCamPlayer(BasePlayer, ThorCamClient):
    """Wrapper for Thor .Net camera based player.
    """

    _config_props_ = (
        'supported_freqs', 'freq', 'supported_taps', 'taps', 'supports_color',
        'exposure_range', 'exposure_ms', 'binning_x_range', 'binning_x',
        'binning_y_range', 'binning_y', 'sensor_size', 'roi_x', 'roi_y',
        'roi_width', 'roi_height', 'gain_range', 'gain', 'black_level_range',
        'black_level', 'frame_queue_size', 'supported_triggers',
        'trigger_type', 'trigger_count', 'num_queued_frames', 'color_gain',
        'serial', 'serials')

    supported_freqs = ListProperty(['20 MHz', ])
    """The supported frequencies."""

    freq = StringProperty('20 MHz')
    """The frequency to use."""

    supported_taps = ListProperty(['1', ])
    """The supported taps."""

    taps = StringProperty('1')
    """The tap to use."""

    supports_color = BooleanProperty(False)
    """Whether the camera supports color."""

    exposure_range = ListProperty([0, 100])
    """The supported exposure range in ms."""

    exposure_ms = NumericProperty(0)
    """The exposure value in ms to use."""

    binning_x_range = ListProperty([0, 0])
    """The supported exposure range."""

    binning_x = NumericProperty(0)
    """The x binning value to use."""

    binning_y_range = ListProperty([0, 0])
    """The supported exposure range."""

    binning_y = NumericProperty(0)
    """The y binning value to use."""

    sensor_size = ListProperty([0, 0])
    """The size of the sensor in pixels."""

    roi_x = NumericProperty(0)
    """The x start position of the ROI in pixels."""

    roi_y = NumericProperty(0)
    """The y start position of the ROI in pixels."""

    roi_width = NumericProperty(0)
    """The width after the x start position of the ROI in pixels, to use."""

    roi_height = NumericProperty(0)
    """The height after the y start position of the ROI in pixels, to use."""

    gain_range = ListProperty([0, 100])
    """The supported exposure range."""

    gain = NumericProperty(0)
    """The gain value to use."""

    black_level_range = ListProperty([0, 100])
    """The supported exposure range."""

    black_level = NumericProperty(0)
    """The black level value to use."""

    frame_queue_size = NumericProperty(1)
    """The max number of image frames to be allowed on the camera's hardware
    queue. Once exceeded, the frames are dropped."""

    supported_triggers = ListProperty(['SW Trigger', 'HW Trigger'])
    """The trigger types supported by the camera."""

    trigger_type = StringProperty('SW Trigger')
    """The trigger type of the camera to use."""

    trigger_count = NumericProperty(0)
    """The number of frames to capture in response to the trigger."""

    num_queued_frames = NumericProperty(0)
    """The number of image frames currently on the camera's hardware queue."""

    color_gain = ListProperty([1, 1, 1])
    """The color gain for each red, green, and blue channel."""

    serials = ListProperty([])
    """The list of serial numbers representing the cameras available.
    """

    serial = StringProperty('')
    """The serial number of the camera that will be opened.
    """

    to_kivy_queue = None
    """The queue that sends messages to be executed in the Kivy thread.
    """

    is_available = BooleanProperty(ThorCamClient is not object)
    """Whether :mod:`thorcam` is available to play."""

    cam_state = StringProperty('none')
    """The current state of the state machine of the camera.

    Can be one of none, opening, open, closing.
    """

    _frame_count = 0

    _ivl_start = 0

    def __init__(self, open_thread=True, **kwargs):
        super(ThorCamPlayer, self).__init__(**kwargs)
        self.can_play = False
        self._kivy_trigger = Clock.create_trigger(self.process_in_kivy_thread)
        self.to_kivy_queue = Queue()
        if ThorCamClient is not object and open_thread:
            self.start_cam_process()

        self.fbind('serial', self._update_summary)
        self._update_summary()

    def _update_summary(self, *largs):
        self.player_summery = 'Thor "{}"'.format(self.serial)

    @error_guard
    def received_camera_response(self, msg, value):
        if msg == 'image':
            value = self.create_image_from_msg(value)
        self.to_kivy_queue.put((msg, value))
        self._kivy_trigger()

    def handle_exception(self, e, exc_info):
        self.to_kivy_queue.put(('exception', (e, exc_info)))
        self._kivy_trigger()

    @error_guard
    def process_in_kivy_thread(self, *largs):
        """Processes messages from the camera in the kivy thread.
        """
        while self.to_kivy_queue is not None:
            try:
                msg, value = self.to_kivy_queue.get(block=False)

                if msg == 'exception':
                    e, exec_info = value
                    cpl_media.error_callback(e, exc_info=exec_info)
                    if self.cam_state == 'open':
                        self.close_camera()
                elif msg == 'cam_open':
                    assert self.cam_state == 'opening'
                    self.cam_state = 'open'
                    self.can_play = True
                elif msg == 'cam_closed':
                    # either remote sent exception, or we asked to stop
                    assert self.cam_state == 'closing'
                    self.cam_state = 'none'
                    self.can_play = False
                    if self.play_state != 'none':
                        self.complete_stop()
                elif msg == 'image':
                    self._handle_image_received(value)
                elif msg == 'playing':
                    if value:
                        assert self.play_state == 'starting'
                    else:
                        assert self.play_state == 'stopping'
                        self.complete_stop()
                elif msg == 'settings':
                    # maintain the last settings
                    old_vals = {
                        key: getattr(self, key) for key in self.settings}
                    for key, val in value.items():
                        setattr(self, key, val)
                        if key in old_vals and old_vals[key] != val and \
                                old_vals[key]:
                            self.set_setting(key, old_vals[key])
                elif msg == 'setting':
                    for key, val in value.items():
                        setattr(self, key, val)
                elif msg == 'serials':
                    self.serials = value
                else:
                    print('Got unknown ThorCamPlayer message', msg, value)
            except Empty:
                break

    def _handle_image_received(self, value):
        """Runs in the kivy thread when we get an image
        """
        t = clock()
        if self.play_state == 'starting':
            self._frame_count = 0
            self.ts_play = self._ivl_start = t

            img = value[0]
            self.metadata_play_used = VideoMetadata(
                img.get_pixel_format(), *img.get_size(), 0)
            self.complete_start()

        if self.play_state != 'playing':
            return

        if t - self._ivl_start >= 1.:
            r = self.real_rate = self._frame_count / (t - self._ivl_start)
            if not self.metadata_play_used.rate:
                self.metadata_play_used = VideoMetadata(
                    *self.metadata_play_used[:3], int(2 * r))
            self._frame_count = 0
            self._ivl_start = t

        self._frame_count += 1
        self.frames_played += 1

        img, count, queued_count, t_img = value
        self.num_queued_frames = queued_count
        self.process_frame(img, {'t': t_img, 'count': count})

    @error_guard
    def open_camera(self, serial):
        """Opens the camera so that we can :meth:`play`.
        """
        if self.cam_state != 'none':
            raise TypeError('Can only open camera if it has not been opened')
        self.cam_state = 'opening'
        self.send_camera_request('open_cam', serial)

    @error_guard
    def close_camera(self):
        """closes the camera and also stops playing if we were playing.
        """
        if self.cam_state != 'open':
            raise TypeError('Can only close camera if it has been opened')
        self.stop()
        self.cam_state = 'closing'
        self.can_play = False
        self.send_camera_request('close_cam', None)

    def refresh_cameras(self):
        """Refreshes the list of Thor cameras plugged in.

        It updates :attr:`serials` when we have the new list.
        """
        self.send_camera_request('serials', None)

    @error_guard
    def play(self):
        if self.cam_state != 'open':
            raise TypeError("Cannot play camera that isn't open")

        self.start_cam_process()
        super(ThorCamPlayer, self).play()
        self.metadata_play_used = VideoMetadata('', 0, 0, 0)
        self.send_camera_request('play')

    def stop(self, *largs, join=False):
        if self.cam_state != 'open':
            return

        if super(ThorCamPlayer, self).stop(join=join):
            self.send_camera_request('stop')

    def set_setting(self, name, value):
        """Requests that the setting should be changed for the camera.

        :param name: The setting name to be changed. E.g. ``"exposure_ms"``.
        :param value: The new setting value.
        """
        self.send_camera_request('setting', (name, value))

    def play_thread_run(self):
        pass

    def stop_all(self, join=False):
        super(ThorCamPlayer, self).stop_all(join=join)
        if ThorCamClient is not object and self._server_thread is not None:
            self.stop_cam_process(join=join)


class ThorCamSettingsWidget(BoxLayout):
    """Settings widget for :class:`ThorCamPlayer`.
    """

    player: ThorCamPlayer = None
    """The player.
    """

    def __init__(self, player=None, **kwargs):
        if player is None:
            player = ThorCamPlayer()
        self.player = player
        super(ThorCamSettingsWidget, self).__init__(**kwargs)


Builder.load_file(join(dirname(__file__), 'thorcam_player.kv'))



"""PTGray based player
======================

This player can play Point Gray ethernet cameras using :mod:`pyflycap2`.
"""

from threading import Thread
from time import perf_counter as clock
from functools import partial
import time
from queue import Queue
from os.path import splitext, join, exists, isdir, abspath, dirname

from ffpyplayer.pic import Image

from kivy.clock import Clock
from kivy.properties import (
    NumericProperty, ReferenceListProperty,
    ObjectProperty, ListProperty, StringProperty, BooleanProperty,
    DictProperty, AliasProperty, OptionProperty, ConfigParserProperty)
from kivy.uix.boxlayout import BoxLayout
from kivy.logger import Logger
from kivy.lang import Builder

from cpl_media.player import BasePlayer, VideoMetadata
from cpl_media import error_guard

try:
    from pyflycap2.interface import GUI, Camera, CameraContext
except ImportError as err:
    GUI = Camera = CameraContext = None
    Logger.debug('cpl_media: Could not import pyflycap2: {}'.format(err))

__all__ = ('PTGrayPlayer', 'PTGraySettingsWidget')


class PTGrayPlayer(BasePlayer):
    """Wrapper for Point Gray based player.
    """

    _config_props_ = (
        'serial', 'ip', 'cam_config_opts', 'brightness', 'exposure',
        'sharpness', 'hue', 'saturation', 'gamma', 'shutter', 'gain',
        'iris', 'frame_rate', 'pan', 'tilt', 'mirror')

    is_available = BooleanProperty(CameraContext is not None)
    """Whether ptgray is available to play."""

    serial = NumericProperty(0)
    '''The serial number of the camera to open. Either :attr:`ip` or
    :attr:`serial` must be provided.
    '''

    serials = ListProperty([])
    """The serials of all the cams available.

    This may only be set by calling :meth:`ask_config`, not set directly.
    """

    ip = StringProperty('')
    '''The IP address of the camera to open. Either :attr:`ip` or
    :attr:`serial` must be provided.
    '''

    ips = ListProperty([])
    """The IPs of all the cams available.

    This may only be set by calling :meth:`ask_config`, not set directly.
    """

    cam_config_opts = DictProperty({'fmt': 'yuv422'})
    '''The configuration options used to configure the camera after opening.
    This are internal and can only be set by the internal thread once
    initially set by config.
    '''

    active_settings = ListProperty([])
    """The list of settings that the camera can control.

    This may only be set by calling :meth:`ask_config`, not set directly.
    """

    brightness = DictProperty({})
    """The camera options for the brightness setting.

    This may only be set by calling :meth:`ask_cam_option_config`, not
    set directly.
    """

    exposure = DictProperty({})
    """The camera options for the exposure setting.

    This may only be set by calling :meth:`ask_cam_option_config`, not
    set directly.
    """

    sharpness = DictProperty({})
    """The camera options for the sharpness setting.

    This may only be set by calling :meth:`ask_cam_option_config`, not
    set directly.
    """

    hue = DictProperty({})
    """The camera options for the hue setting.

    This may only be set by calling :meth:`ask_cam_option_config`, not
    set directly.
    """

    saturation = DictProperty({})
    """The camera options for the saturation setting.

    This may only be set by calling :meth:`ask_cam_option_config`, not
    set directly.
    """

    gamma = DictProperty({})
    """The camera options for the gamma setting.

    This may only be set by calling :meth:`ask_cam_option_config`, not
    set directly.
    """

    shutter = DictProperty({})
    """The camera options for the shutter setting.

    This may only be set by calling :meth:`ask_cam_option_config`, not
    set directly.
    """

    gain = DictProperty({})
    """The camera options for the gain setting.

    This may only be set by calling :meth:`ask_cam_option_config`, not
    set directly.
    """

    iris = DictProperty({})
    """The camera options for the iris setting.

    This may only be set by calling :meth:`ask_cam_option_config`, not
    set directly.
    """

    frame_rate = DictProperty({})
    """The camera options for the frame_rate setting.

    This may only be set by calling :meth:`ask_cam_option_config`, not
    set directly.
    """

    pan = DictProperty({})
    """The camera options for the pan setting.

    This may only be set by calling :meth:`ask_cam_option_config`, not
    set directly.
    """

    tilt = DictProperty({})
    """The camera options for the tilt setting.

    This may only be set by calling :meth:`ask_cam_option_config`, not
    set directly.
    """

    mirror = BooleanProperty(False)
    """Whether the camera is mirrored. Read only.
    """

    config_thread = None
    """The configuration thread.
    """

    config_queue = None
    """The configuration queue.
    """

    config_active_queue = ListProperty([])
    """The items currently being configured in the configuration thread.
    """

    config_active = BooleanProperty(False)
    """Whether the configuration is currently active.
    """

    _camera = None

    ffmpeg_pix_map = {
        'mono8': 'gray', 'yuv411': 'uyyvyy411', 'yuv422': 'uyvy422',
        'yuv444': 'yuv444p', 'rgb8': 'rgb8', 'mono16': 'gray16le',
        'rgb16': 'rgb565le', 's_mono16': 'gray16le', 's_rgb16': 'rgb565le',
        'bgr': 'bgr24', 'bgru': 'bgra', 'rgb': 'rgb24', 'rgbu': 'rgba',
        'bgr16': 'bgr565le', 'yuv422_jpeg': 'yuvj422p'}
    """Pixel formats supported by the camera and their :mod:`ffpyplayer`
    equivalent.
    """

    def __init__(self, open_thread=True, **kwargs):
        self.active_settings = self.get_setting_names()
        super(PTGrayPlayer, self).__init__(**kwargs)
        if CameraContext is not None and open_thread:
            self.start_config()

        def do_serial(*largs):
            self.ask_config('serial')

        self.fbind('serial', do_serial)

        def do_ip(*largs):
            self.ask_config('serial')

        self.fbind('ip', do_ip)
        do_ip()

        self.fbind('ip', self._update_summary)
        self.fbind('serial', self._update_summary)
        self._update_summary()

    def _update_summary(self, *largs):
        name = str(self.serial or self.ip)
        self.player_summery = 'PTGray "{}"'.format(name)

    def start_config(self, *largs):
        """Called by `__init__` to start the configuration thread.
        """
        self.config_queue = Queue()
        self.config_active_queue = []
        thread = self.config_thread = Thread(
            target=self.config_thread_run, name='Config thread')
        thread.start()
        self.ask_config('serials')

    @error_guard
    def stop_config(self, *largs, join=False):
        """Stops the configuration thread.
        """
        self.ask_config('eof', ignore_play=True)
        if join and self.config_thread:
            self.config_thread.join()
            self.config_thread = None

    @error_guard
    def ask_config(self, item, ignore_play=False):
        """Asks to read the config values of the item. E.g. ``'serials'``
        for the list of serials or ``'gui'`` to show the PTGray GUI to the
        user.

        :param item: The request to send.
        :param ignore_play: Whether to send it event if the camera is playing.
        """
        if not ignore_play and self.play_state != 'none':
            raise TypeError('Cannot configure while playing')

        queue = self.config_queue
        if queue is not None:
            self.config_active = True
            if item != 'eof':
                # XXX: really strange bug, but somehow if this is set here when
                # we call stop_all and we join, it blocks forever on setting
                # can_play. Blocking only happens when kv binds to can_play.
                # Makes no sense as it's all from the same thread???
                self.can_play = False
            self.config_active_queue.append(item)
            queue.put(item)

    def ask_cam_option_config(self, setting, name, value):
        """Asks to set the setting of the camera to a specific value.

        :param setting: The setting, e.g. ``"brightness"``.
        :param name: How to set it, e.g. ``"value"`` to set the value or
            ``'one push'`` to auto configure it.
        :param value: The value to use to set it to.
        """
        if not name or getattr(self, setting)[name] != value:
            self.ask_config(
                ('option', (setting, name, value)), ignore_play=True)

    def finish_ask_config(self, item, *largs, **kwargs):
        """Called in the kivy thread automatically after the camera
        has been re-configured.
        """
        if isinstance(item, tuple) and item[0] == 'option':
            setting, _, _ = item[1]
            getattr(self, setting).update(kwargs['values'])
            set_rate = setting in ('frame_rate', 'metadata_play_used')
        else:
            for k, v in kwargs.items():
                setattr(self, k, v)
            set_rate = 'frame_rate' in kwargs or 'metadata_play_used' in kwargs

        self.active_settings = self.get_active_settings()

        if set_rate:
            fmt, w, h, r = self.metadata_play_used
            if 'max' in self.frame_rate:
                r = max(r, self.frame_rate['max'])
                self.metadata_play_used = VideoMetadata(fmt, w, h, r)

    @error_guard
    def _remove_config_item(self, item, *largs):
        self.config_active_queue.remove(item)
        if not self.config_active_queue:
            self.config_active = False
            self.can_play = True

    def get_active_settings(self):
        """List of settings supported by the camera.
        """
        settings = []
        for setting in self.get_setting_names():
            if getattr(self, setting).get('present', False):
                settings.append(setting)
        return list(sorted(settings))

    def get_setting_names(self):
        """List of all settings potentially supported by a camera.
        """
        return list(sorted((
            'brightness', 'exposure', 'sharpness', 'hue', 'saturation',
            'gamma', 'shutter', 'gain', 'iris', 'frame_rate', 'pan', 'tilt')))

    def read_cam_option_config(self, setting, cam):
        """Reads the setting from the camera.

        Called from the internal configuration thread.
        """
        options = {}
        mn, mx = cam.get_cam_abs_setting_range(setting)
        options['min'], options['max'] = mn, mx
        options['value'] = cam.get_cam_abs_setting_value(setting)
        options.update(cam.get_cam_setting_option_values(setting))
        return options

    def write_cam_option_config(self, setting, cam, name, value):
        """Writes the setting to the camera.

        Called from the internal configuration thread.
        """
        if name == 'value':
            cam.set_cam_abs_setting_value(setting, value)
        else:
            cam.set_cam_setting_option_values(setting, **{name: value})
            if name == 'one_push' and value:
                while cam.get_cam_setting_option_values(setting)['one_push']:
                    time.sleep(.2)

    def write_cam_options_config(self, cam):
        """Writes all the settings as provided as properties of this instance
        to the camera.

        Called from the internal configuration thread.
        """
        for setting in self.get_setting_names():
            settings = getattr(self, setting)
            cam.set_cam_setting_option_values(
                setting, abs=settings.get('abs', None),
                controllable=settings.get('controllable', None),
                auto=settings.get('auto', None)
            )
            settings_read = cam.get_cam_setting_option_values(setting)
            if settings_read['controllable'] and not settings_read['auto']:
                if settings_read['abs'] and 'value' in settings:
                    cam.set_cam_abs_setting_value(setting, settings['value'])
                elif not settings_read['abs'] and 'relative_value' in settings:
                    cam.set_cam_setting_option_values(
                        setting, relative_value=settings['relative_value'])

        if cam.get_horizontal_mirror()[0]:
            cam.set_horizontal_mirror(self.mirror)

    def read_cam_options_config(self, cam):
        """Reads all the settings of this instance
        to the camera.

        Called from the internal configuration thread.
        """
        for setting in self.get_setting_names():
            Clock.schedule_once(partial(
                self.finish_ask_config, None,
                **{setting: self.read_cam_option_config(setting, cam)}))

        if cam.get_horizontal_mirror()[0]:
            Clock.schedule_once(partial(
                self.finish_ask_config, None,
                mirror=cam.get_horizontal_mirror()[1]))

    def write_gige_opts(self, c, opts):
        """Writes the GIGE setting to the camera.

        Called from the internal configuration thread.
        """
        c.set_gige_mode(opts['mode'])
        c.set_drop_mode(opts['drop'])
        c.set_gige_config(opts['offset_x'], opts['offset_y'], opts['width'],
                          opts['height'], opts['fmt'])
        c.set_gige_packet_config(opts['resend'], opts['timeout'],
                                 opts['timeout_retries'])
        c.set_gige_binning(opts['horizontal'], opts['vertical'])

    def read_gige_opts(self, c):
        """Reads the GIGE setting from the camera.

        Called from the internal configuration thread.
        """
        opts = self.cam_config_opts
        opts['drop'] = c.get_drop_mode()
        opts.update(c.get_gige_config())
        opts['mode'] = c.get_gige_mode()
        opts.update(c.get_gige_packet_config())
        opts['horizontal'], opts['vertical'] = c.get_gige_binning()

    def config_thread_run(self):
        """The function run by the configuration thread.
        """
        queue = self.config_queue
        cc = CameraContext()

        while True:
            item = queue.get()
            try:
                if item == 'eof':
                    return

                ip = ''
                serial = 0
                do_serial = False
                if item == 'serials':
                    cc.rescan_bus()
                    cams = cc.get_gige_cams()
                    old_serial = serial = self.serial
                    old_ip = ip = self.ip

                    ips = ['.'.join(map(str, Camera(serial=s).ip))
                           for s in cams]
                    if cams:
                        if serial not in cams and ip not in ips:
                            serial = 0
                            ip = ''
                        elif serial in cams:
                            ip = ips[cams.index(serial)]
                        else:
                            serial = cams[ips.index(ip)]

                    Clock.schedule_once(partial(
                        self.finish_ask_config, item, serials=cams,
                        serial=serial, ips=ips, ip=ip))

                    if serial:
                        c = Camera(serial=serial)
                        c.connect()
                        if old_serial == serial or old_ip == ip:
                            self.write_gige_opts(c, self.cam_config_opts)
                            self.write_cam_options_config(c)
                        self.read_gige_opts(c)
                        self.read_cam_options_config(c)

                        if self.cam_config_opts['fmt'] not in \
                            self.ffmpeg_pix_map or \
                                self.cam_config_opts['fmt'] == 'yuv411':
                            self.cam_config_opts['fmt'] = 'rgb'
                            self.write_gige_opts(c, self.cam_config_opts)
                        c.disconnect()
                        c = None
                elif item == 'serial':
                    do_serial = True
                elif item == 'gui':
                    gui = GUI()
                    gui.show_selection()
                    do_serial = True  # read possibly updated config
                elif c or self._camera:
                    cam = c or self._camera
                    if isinstance(item, tuple) and item[0] == 'mirror':
                        if cam.get_horizontal_mirror()[0]:
                            cam.set_horizontal_mirror(item[1])
                        Clock.schedule_once(partial(
                            self.finish_ask_config, item,
                            mirror=cam.get_horizontal_mirror()[1]))
                    elif isinstance(item, tuple) and item[0] == 'option':
                        _, (setting, name, value) = item
                        if name:
                            self.write_cam_option_config(
                                setting, cam, name, value)
                        Clock.schedule_once(partial(
                            self.finish_ask_config, item,
                            values=self.read_cam_option_config(setting, cam)))

                if do_serial:
                    _ip = ip = self.ip
                    serial = self.serial
                    if serial or ip:
                        if _ip:
                            _ip = list(map(int, _ip.split('.')))
                        c = Camera(serial=serial or None, ip=_ip or None)
                        serial = c.serial
                        ip = '.'.join(map(str, c.ip))
                        c.connect()
                        self.read_gige_opts(c)
                        self.read_cam_options_config(c)

                        if self.cam_config_opts['fmt'] not in \
                            self.ffmpeg_pix_map or \
                                self.cam_config_opts['fmt'] == 'yuv411':
                            self.cam_config_opts['fmt'] = 'rgb'
                            self.write_gige_opts(c, self.cam_config_opts)
                        c.disconnect()
                        c = None

                if serial or ip:
                    opts = self.cam_config_opts
                    if opts['fmt'] not in self.ffmpeg_pix_map:
                        raise Exception('Pixel format {} cannot be converted'.
                                        format(opts['fmt']))
                    if opts['fmt'] == 'yuv411':
                        raise ValueError('yuv411 is not currently supported')
                    metadata = VideoMetadata(
                        self.ffmpeg_pix_map[opts['fmt']], opts['width'],
                        opts['height'], 30.0)
                    Clock.schedule_once(partial(
                        self.finish_ask_config, item, metadata_play=metadata,
                        metadata_play_used=metadata, serial=serial, ip=ip))
            except Exception as e:
                self.exception(e)
            finally:
                Clock.schedule_once(partial(self._remove_config_item, item))

    def play_thread_run(self):
        process_frame = self.process_frame
        c = None
        ffmpeg_fmts = self.ffmpeg_pix_map

        try:
            ip = list(map(int, self.ip.split('.'))) if self.ip else None
            c = Camera(serial=self.serial or None, ip=ip)
            c.connect()

            started = False
            # use_rt = self.use_real_time
            count = 0
            ivl_start = 0

            c.start_capture()
            while self.play_state != 'stopping':
                try:
                    c.read_next_image()
                except Exception as e:
                    self.exception(e)
                    continue
                if not started:
                    ivl_start = clock()
                    self.setattr_in_kivy_thread('ts_play', ivl_start)
                    Clock.schedule_once(self.complete_start)
                    started = True
                    self._camera = c

                ivl_end = clock()
                if ivl_end - ivl_start >= 1.:
                    real_rate = count / (ivl_end - ivl_start)
                    self.setattr_in_kivy_thread('real_rate', real_rate)
                    count = 0
                    ivl_start = ivl_end

                count += 1
                self.increment_in_kivy_thread('frames_played')

                image = c.get_current_image()
                pix_fmt = image['pix_fmt']
                if pix_fmt not in ffmpeg_fmts:
                    raise Exception('Pixel format {} cannot be converted'.
                                    format(pix_fmt))
                ff_fmt = ffmpeg_fmts[pix_fmt]
                if ff_fmt == 'yuv444p':
                    buff = image['buffer']
                    img = Image(
                        plane_buffers=[buff[1::3], buff[0::3], buff[2::3]],
                        pix_fmt=ff_fmt, size=(image['cols'], image['rows']))
                elif pix_fmt == 'yuv411':
                    raise ValueError('yuv411 is not currently supported')
                else:
                    img = Image(
                        plane_buffers=[image['buffer']], pix_fmt=ff_fmt,
                        size=(image['cols'], image['rows']))

                process_frame(img, {'t': ivl_end})
        except Exception as e:
            self.exception(e)
        finally:
            self._camera = None

            try:
                c.disconnect()
            finally:
                Clock.schedule_once(self.complete_stop)

    def stop_all(self, join=False):
        self.stop_config(join=join)
        super(PTGrayPlayer, self).stop_all(join=join)


class PTGraySettingsWidget(BoxLayout):
    """Settings widget for :class:`PTGrayPlayer`.
    """

    settings_last = ''
    """The last name of the setting the GUI currently controls (i.e. one
    of :meth:`PTGrayPlayer.get_active_settings`).
    """

    opt_settings = DictProperty({})
    """The values for the settings currently being controlled in the GUI.
    """

    player: PTGrayPlayer = None
    """The player.
    """

    def __init__(self, player=None, **kwargs):
        if player is None:
            player = PTGrayPlayer()
        self.player = player
        super(PTGraySettingsWidget, self).__init__(**kwargs)

    def _track_setting(self, *largs):
        self.opt_settings = getattr(self.player, self.settings_last)

    def bind_pt_setting(self, setting):
        """Tracks the setting currently selected in the GUI and auto-updates
        when it changes.
        """
        if self.settings_last:
            self.player.funbind(self.settings_last, self._track_setting)
        self.settings_last = ''
        self.opt_settings = {}

        if setting:
            self.settings_last = setting
            self.player.fbind(setting, self._track_setting)
            self._track_setting()


Builder.load_file(join(dirname(__file__), 'ptgray_player.kv'))



import os
import trio

from kivy.config import Config
Config.set('graphics', 'width', '1600')
Config.set('graphics', 'height', '900')
for items in Config.items('input'):
    Config.remove_option('input', items[0])


from cpl_media.tests.app.demo_app import DemoApp
from kivy.tests.async_common import UnitKivyApp


class DemoTestApp(UnitKivyApp, DemoApp):

    _images_showed = 0

    def __init__(self, ini_file, **kwargs):
        self._ini_config_filename = ini_file
        self._data_path = os.path.dirname(ini_file)
        super(DemoTestApp, self).__init__(**kwargs)

    def check_close(self):
        super(DemoTestApp, self).check_close()
        return True

    def handle_exception(self, msg, exc_info=None,
                         level='error', *largs):
        super(DemoTestApp, self).handle_exception(msg, exc_info, level, *largs)

        if isinstance(exc_info, str):
            self.get_logger().error(msg)
            self.get_logger().error(exc_info)
        elif exc_info is not None:
            tp, value, tb = exc_info
            try:
                if value is None:
                    value = tp()
                if value.__traceback__ is not tb:
                    raise value.with_traceback(tb)
                raise value
            finally:
                value = None
                tb = None
        elif level in ('error', 'exception'):
            raise Exception(msg)

    def _display_frame(self, image, metadata):
        super(DemoTestApp, self)._display_frame(image, metadata)
        self._images_showed += 1



"""Remote server recorder
=========================

This class acts as a recorder that receives media from a camera and records it
to the network. E.g. the server can be configured
to receive video from a FFmpeg player and send it to the network where a client
player plays the video.
"""
from itertools import accumulate
from threading import Thread
import socket
import sys
import struct
from time import perf_counter as clock
from queue import Queue, Empty
import traceback
from os.path import splitext, join, exists, isdir, abspath, dirname
import select

from kivy.properties import ObjectProperty, NumericProperty, StringProperty, \
    BooleanProperty, ListProperty
from kivy.logger import Logger
from kivy.clock import Clock
from kivy.uix.boxlayout import BoxLayout
from kivy.lang import Builder

from more_kivy_app.utils import yaml_dumps, yaml_loads

from cpl_media import error_guard
from cpl_media.recorder import BaseRecorder
import cpl_media

__all__ = ('RemoteVideoRecorder', 'RemoteRecordSettingsWidget', 'RemoteData',
           'EndConnection')


class EndConnection(Exception):
    """Raised when the connection is closed.
    """
    pass


connection_errors = (
    EndConnection, ConnectionAbortedError, ConnectionResetError)


class RemoteData(object):
    """Provides methods to send and receive messages from a socket.

    """

    def send_msg(self, sock, msg, value):
        """Sends message to the server.

        :param sock: The socket
        :param msg: The message name string (e.g. image).
        :param value: The message value.
        :return:
        """
        # ts = time.monotonic()
        if msg == 'image':
            image, metadata = value
            bin_data = image.to_bytearray()
            data = yaml_dumps((
                'image', (list(map(len, bin_data)), image.get_pixel_format(),
                          image.get_size(), image.get_linesizes(), metadata)))
            data = data.encode('utf8')
        else:
            data = yaml_dumps((msg, value))
            data = data.encode('utf8')
            bin_data = []
        # ts2 = time.monotonic()

        sock.sendall(struct.pack('>II', len(data), sum(map(len, bin_data))))
        sock.sendall(data)
        for item in bin_data:
            sock.sendall(item)
        # print('part1: {}, part2: {}'.format(ts2 - ts, time.monotonic() - ts2))

    def decode_data(self, msg_buff, msg_len):
        """Decodes buffer data received from the network.

        :param msg_buff: The bytes data received so far.
        :param msg_len: The expected size of the message as tuple -
            The size of the message and any associated binary data.
        :return: A tuple of the message name and value, or (None, None) if
            we haven't read the full message.
        """
        n, bin_n = msg_len
        assert n + bin_n == len(msg_buff)
        data = msg_buff[:n].decode('utf8')
        msg, value = yaml_loads(data)

        if msg == 'image':
            bin_data = msg_buff[n:]
            planes_sizes, pix_fmt, size, linesize, metadata = value
            starts = list(accumulate([0] + list(planes_sizes[:-1])))
            ends = accumulate(planes_sizes)
            planes = [bin_data[s:e] for s, e in zip(starts, ends)]

            value = planes, pix_fmt, size, linesize, metadata
        else:
            assert not bin_n
        return msg, value

    def read_msg(self, sock, msg_len, msg_buff):
        """Reads the message and decodes it once we read the full message.

        :param sock: The socket.
        :param msg_len: The tuple of the message length and associated data.
            If empty, the start of the next message will provide this
            information.
        :param msg_buff: The message buffer.
        :return: A 4-tuple of ``(msg_len, msg_buff, msg, value)``. Where
            ``msg_len, msg_buff`` are similar to the input, and
            ``(msg, value)`` is the message and its value if we read a full
            message, otherwise they are None.
        """
        # still reading msg size
        msg = value = None
        if not msg_len:
            assert 8 - len(msg_buff)
            data = sock.recv(8 - len(msg_buff))
            if not data:
                raise EndConnection('Remote client was closed')

            msg_buff += data
            if len(msg_buff) == 8:
                msg_len = struct.unpack('>II', msg_buff)
                msg_buff = b''
        else:
            total = sum(msg_len)
            assert total - len(msg_buff)
            data = sock.recv(total - len(msg_buff))
            if not data:
                raise EndConnection('Remote client was closed')

            msg_buff += data
            if len(msg_buff) == total:
                msg, value = self.decode_data(msg_buff, msg_len)

                msg_len = ()
                msg_buff = b''
        return msg_len, msg_buff, msg, value


class RemoteVideoRecorder(BaseRecorder, RemoteData):
    """A server recorder that takes images from a player and sends it
    to a client player over a socket.

    The server is intended to be started before anything else can be
    processed. Once the server is run, we can start recording
    (whether a client is connected or not). If the client is not playing,
    we don't send frames and just skip them.

    A client should accept these messages: exception, started_recording,
    stopped_recording, or image.
    Server accepts messages from client: started_playing, stopped_playing.

    We send started_recording in response to started_playing request.
    Either immediately if we were already recording, or later when we start.
    We send stopped_recording in response to stopping recording, if the client
    was between started_playing and stopped_playing (i.e. it expected to get
    images).
    We send images if between started_playing and stopped_playing requests
    and if we are recording.

    We only accept started_playing either before any started_playing
    message or after a stopped_playing message (i.e. no duplicates).
    """

    _config_props_ = ('server', 'port', 'timeout', 'max_images_buffered')

    server = StringProperty('localhost')
    """The server address on which to broadcast the data.
    """

    port = NumericProperty(10000)
    """The server port on which to broadcast the data.
    """

    timeout = NumericProperty(.01)
    """How long to wait before timing out when reading data before checking the
    queue for other requests.
    """

    server_active = BooleanProperty(False)
    """Whether the server is currently running.
    """

    max_images_buffered = NumericProperty(5)
    """How many images the server should buffer before it starts dropping
    images, rather than queuing them to be sent to the client.
    """

    from_kivy_queue = None
    """The queue that receives messages from Kivy.

    This queue can receive these messages: eof, image, started_recording,
    or stopped_recording.
    """

    to_kivy_queue = None
    """The queue that sends messages to Kivy.
    """

    _kivy_trigger = None
    """Trigger for kivy thread to read the queue - to be called after adding
    something to the queue.
    """

    server_thread = None
    """The server thread instance.
    """

    _server_client_playing = False
    """Whether the client is playing

    May only be set from the server thread.
    """

    _server_client_requested_playing = False
    """Whether the client requested playing

    May only be set from the server thread.
    """

    _server_recording = None
    """Keeps track of whether we are recording rn

    May only be set from the server thread.
    """

    _first_image_while_playing = False
    """If this is the first image right after we started recording.

    May only be set from the server thread.
    """

    def __init__(self, **kwargs):
        super(RemoteVideoRecorder, self).__init__(**kwargs)
        self._kivy_trigger = Clock.create_trigger(self.process_in_kivy_thread)

        self.fbind('server', self._update_summary)
        self.fbind('port', self._update_summary)
        self._update_summary()

    def _update_summary(self, *largs):
        self.recorder_summery = 'Network "{}:{}"'.format(
            self.server, self.port)

    def server_run(self, from_kivy_queue, to_kivy_queue):
        """Server method, that is executed in the internal server thread.
        """
        trigger = self._kivy_trigger
        timeout = self.timeout

        # Create a TCP/IP socket
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

        # Connect the socket to the port where the server is listening
        server_address = (self.server, self.port)
        Logger.info('RemoteVideoRecorder: starting up on {} port {}'
                    .format(*server_address))

        try:
            sock.bind(server_address)
            sock.listen(1)

            while True:
                r, _, _ = select.select([sock], [], [], timeout)
                if not r:
                    try:
                        while True:
                            if 'eof' == self._server_message_from_queue(
                                    None, from_kivy_queue):
                                return
                    except Empty:
                        pass

                    continue

                connection, client_address = sock.accept()
                msg_len, msg_buff = (), b''

                try:
                    while True:
                        r, _, _ = select.select([connection], [], [], timeout)
                        if r:
                            msg_len, msg_buff, msg, value = self.read_msg(
                                connection, msg_len, msg_buff)
                            if msg is not None:
                                if not self._server_message_from_client(
                                        connection, msg, value):
                                    to_kivy_queue.put((msg, value))
                                    trigger()

                        try:
                            while True:
                                if 'eof' == self._server_message_from_queue(
                                        connection, from_kivy_queue):
                                    return
                        except Empty:
                            pass
                except connection_errors:
                    pass
                finally:
                    Logger.info(
                        'RemoteVideoRecorder: closing client connection')
                    self._server_client_playing = False
                    self._server_client_requested_playing = False
                    try:
                        connection.close()
                    except Exception:
                        pass
        except Exception as e:
            exc_info = ''.join(traceback.format_exception(*sys.exc_info()))
            to_kivy_queue.put(
                ('exception', (str(e), exc_info)))
            trigger()
        finally:
            Logger.info('closing socket')
            sock.close()

    def _server_message_from_client(self, connection, msg, value):
        """Processes message from the client.

        Executed from the internal thread.
        """
        try:
            if msg == 'started_playing':
                if self._server_client_playing or \
                        self._server_client_requested_playing:
                    raise TypeError(
                        'Client already notified that it is playing')

                if self._server_recording:
                    self._server_client_playing = True
                    self.send_msg(connection, 'started_recording',
                                  self._server_recording)
                else:
                    self._server_client_requested_playing = True
                return True
            elif msg == 'stopped_playing':
                self._server_client_requested_playing = False
                self._server_client_playing = False
                self.send_msg(connection, 'stopped_playing', None)
                return True
        except Exception as e:
            exc_info = ''.join(traceback.format_exception(*sys.exc_info()))
            self.send_msg(connection, 'exception', (str(e), exc_info))
            return True
        return False

    def _server_message_from_queue(self, connection, from_kivy_queue):
        """Processes message from the kivy queue.

        Executed from the internal thread.
        """
        msg, value = from_kivy_queue.get_nowait()
        if msg == 'eof':
            return 'eof'

        if msg == 'image':
            if self._first_image_while_playing:
                self.setattr_in_kivy_thread('ts_record', clock())
                self._first_image_while_playing = False

            if self._server_client_playing:
                assert connection is not None
                self.increment_in_kivy_thread(
                    'size_recorded', sum(value[0].get_buffer_size()))
                self.increment_in_kivy_thread('frames_recorded')

                self.send_msg(connection, msg, value)
            else:
                self.increment_in_kivy_thread('frames_skipped')
        elif msg == 'started_recording':
            self._server_recording = recording = tuple(value)
            # cannot be playing as it should at most be in requested_playing
            assert not self._server_client_playing
            self._first_image_while_playing = True

            if self._server_client_requested_playing:
                assert connection is not None
                self._server_client_requested_playing = False
                self._server_client_playing = True
                self.send_msg(connection, 'started_recording', recording)
        elif msg == 'stopped_recording':
            assert self._server_recording is not None
            self._server_recording = None
            if self._server_client_requested_playing or \
                    self._server_client_playing:
                assert connection is not None
                self.send_msg(connection, 'stopped_recording', None)
                self._server_client_requested_playing = False
                self._server_client_playing = False
        elif connection is not None:
            self.send_msg(connection, msg, value)

    @error_guard
    def send_message_to_client(self, msg, value):
        """Sends the message to the client through the server.

        :param msg: The message name string.
        :param value: The message value.
        """
        if self.from_kivy_queue is None:
            return

        self.from_kivy_queue.put((msg, value))

    @error_guard
    def send_image_to_client(self, image):
        """Sends a image (tuple of image, metadata) to the client through the
        server.
        """
        if self.from_kivy_queue is None:
            return
        image, metadata = image

        if not self.max_images_buffered or \
                self.from_kivy_queue.qsize() < self.max_images_buffered:
            self.from_kivy_queue.put(('image', (image, metadata)))
        else:
            self.increment_in_kivy_thread('frames_skipped')

    @error_guard
    def process_in_kivy_thread(self, *largs):
        """Processes messages from the server in the kivy thread.
        """
        while self.to_kivy_queue is not None:
            try:
                msg, value = self.to_kivy_queue.get(block=False)

                if msg == 'exception':
                    e, exec_info = value
                    cpl_media.error_callback(e, exc_info=exec_info)
                    self.stop_server()
                else:
                    print('Got unknown RemoteVideoRecorder message',
                          msg, value)
            except Empty:
                break

    def start_server(self):
        """Starts the server, so that we can :meth:`record`.
        """
        if self.server_thread is not None:
            return

        self.server_active = True
        self._server_client_playing = False
        self._server_client_requested_playing = False
        self._server_recording = None
        from_kivy_queue = self.from_kivy_queue = Queue()
        to_kivy_queue = self.to_kivy_queue = Queue()

        server_thread = self.server_thread = Thread(
            target=self.server_run, args=(from_kivy_queue, to_kivy_queue))
        server_thread.start()

    @error_guard
    def stop_server(self, join=False):
        """Stops the server and also stops recording if we were recording.
        """
        self.stop(join=join)
        if self.server_thread is None:
            return

        self.from_kivy_queue.put(('eof', None))
        if join:
            self.server_thread.join()

        self.server_thread = self.to_kivy_queue = self.from_kivy_queue = None
        self.server_active = False

    @error_guard
    def record(self, *largs, **kwargs):
        self.start_server()
        super(RemoteVideoRecorder, self).record(*largs, **kwargs)

        self.metadata_record_used = self.metadata_player
        self.player.frame_callbacks.append(self.send_image_to_client)
        self.from_kivy_queue.put(
            ('started_recording', self.metadata_record_used))

        self.complete_start()

    @error_guard
    def stop(self, *largs, join=False):
        if super(RemoteVideoRecorder, self).stop(join=join):
            self.player.frame_callbacks.remove(self.send_image_to_client)
            self.from_kivy_queue.put(('stopped_recording', None))
            self.complete_stop()

    def stop_all(self, join=False):
        super(RemoteVideoRecorder, self).stop_all(join=join)
        self.stop_server(join=join)

    def record_thread_run(self, *largs):
        pass


class RemoteRecordSettingsWidget(BoxLayout):
    """Settings widget for :class:`RemoteVideoRecorder`.
    """

    recorder: RemoteVideoRecorder = None
    """The recorder.
    """

    def __init__(self, recorder=None, **kwargs):
        if recorder is None:
            recorder = RemoteVideoRecorder()
        self.recorder = recorder
        super(RemoteRecordSettingsWidget, self).__init__(**kwargs)


Builder.load_file(join(dirname(__file__), 'server_recorder.kv'))



"""Remote player
================

This class plays video from the network. E.g. a
:class:`cpl_media.remote.server.RemoteData` can be configured
to broadcast video from a FFmpeg camera. This client plays the video from
the server.
"""

from threading import Thread
from time import perf_counter as clock
import socket
import sys
from queue import Queue, Empty
from os.path import splitext, join, exists, isdir, abspath, dirname
import traceback
import select

from ffpyplayer.pic import Image, SWScale

from kivy.logger import Logger
from kivy.properties import StringProperty, NumericProperty, BooleanProperty
from kivy.clock import Clock
from kivy.uix.boxlayout import BoxLayout
from kivy.lang import Builder

from cpl_media import error_guard
from cpl_media.player import BasePlayer, VideoMetadata
import cpl_media
from .server import RemoteData

__all__ = ('RemoteVideoPlayer', 'ClientPlayerSettingsWidget')


class RemoteVideoPlayer(BasePlayer, RemoteData):
    """A player that is a network client that plays images received from a
    :class:`cpl_media.remote.server.RemoteData` over the network.
    """

    _config_props_ = ('server', 'port', 'timeout')

    server = StringProperty('')
    """The server address that broadcasts the data.
    """

    port = NumericProperty(0)
    """The server port that broadcasts the data.
    """

    timeout = NumericProperty(.01)
    """How long to wait before timing out when reading data before checking the
    queue for other requests.
    """

    client_active = BooleanProperty(False)
    """Whether the client thread is currently running.
    """

    from_kivy_queue = None
    """The queue that receives messages from Kivy.
    """

    to_kivy_queue = None
    """The queue that sends messages to Kivy.
    """

    _kivy_trigger = None
    """Trigger for kivy thread to read the queue - to be called after adding
    something to the queue.
    """

    listener_thread = None
    """The client thread instance.
    """

    _frame_count = 0

    _ivl_start = 0

    def __init__(self, **kwargs):
        super(RemoteVideoPlayer, self).__init__(**kwargs)
        self._kivy_trigger = Clock.create_trigger(self.process_in_kivy_thread)

        self.fbind('server', self._update_summary)
        self.fbind('port', self._update_summary)
        self._update_summary()

    def _update_summary(self, *largs):
        self.player_summery = 'Network "{}:{}"'.format(self.server, self.port)

    def listener_run(self, from_kivy_queue, to_kivy_queue):
        """Client method, that is executed in the internal client thread.
        """
        trigger = self._kivy_trigger
        timeout = self.timeout

        # Create a TCP/IP socket
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.settimeout(2)

        # Connect the socket to the port where the server is listening
        server_address = (self.server, self.port)
        Logger.info('RemoteVideoPlayer: connecting to {} port {}'
                    .format(*server_address))

        msg_len, msg_buff = (), b''

        try:
            sock.connect(server_address)
            done = False

            while not done:
                r, _, _ = select.select([sock], [], [], timeout)
                if r:
                    msg_len, msg_buff, msg, value = self.read_msg(
                        sock, msg_len, msg_buff)
                    if msg is not None:
                        to_kivy_queue.put((msg, value))
                        trigger()

                try:
                    while True:
                        msg, value = from_kivy_queue.get_nowait()
                        if msg == 'eof':
                            done = True
                            break
                        else:
                            self.send_msg(sock, msg, value)
                except Empty:
                    pass
        except Exception as e:
            exc_info = ''.join(traceback.format_exception(*sys.exc_info()))
            to_kivy_queue.put(
                ('exception_exit', (str(e), exc_info)))
            trigger()
        finally:
            Logger.info('RemoteVideoPlayer: closing socket')
            sock.close()

    def send_message_to_server(self, key, value):
        """Sends the message to the server over the network.

        :param msg: The message name string.
        :param value: The message value.
        """
        if self.from_kivy_queue is None:
            return
        self.from_kivy_queue.put((key, value))

    def start_listener(self):
        """Starts the client, so that we can :meth:`play`.
        """
        if self.listener_thread is not None:
            return

        self.client_active = True
        from_kivy_queue = self.from_kivy_queue = Queue()
        to_kivy_queue = self.to_kivy_queue = Queue()
        thread = self.listener_thread = Thread(
            target=self.listener_run, args=(from_kivy_queue, to_kivy_queue))
        thread.start()

    @error_guard
    def process_in_kivy_thread(self, *largs):
        """Processes messages from the client in the kivy thread.
        """
        while self.to_kivy_queue is not None:
            try:
                msg, value = self.to_kivy_queue.get(block=False)

                if msg == 'exception':
                    e, exec_info = value
                    cpl_media.error_callback(e, exc_info=exec_info)
                elif msg == 'exception_exit':
                    e, exec_info = value
                    cpl_media.error_callback(e, exc_info=exec_info)
                    self.stop_all()
                    if self.play_state != 'none':
                        self.complete_stop()
                elif msg == 'started_recording':
                    if self.play_state == 'starting':
                        self.ts_play = self._ivl_start = clock()
                        self._frame_count = 0

                        self.metadata_play_used = VideoMetadata(*value)
                        self.complete_start()
                elif msg == 'stopped_recording':
                    self.stop()
                elif msg == 'stopped_playing':
                    self.complete_stop()
                elif msg == 'image':
                    if self.play_state != 'playing':
                        continue

                    t = clock()
                    if t - self._ivl_start >= 1.:
                        self.real_rate = self._frame_count / (
                            t - self._ivl_start)
                        self._frame_count = 0
                        self._ivl_start = t

                    self._frame_count += 1
                    self.frames_played += 1

                    plane_buffers, pix_fmt, size, linesize, metadata = value
                    sws = SWScale(*size, pix_fmt, ofmt=pix_fmt)
                    img = Image(
                        plane_buffers=plane_buffers, pix_fmt=pix_fmt,
                        size=size, linesize=linesize)
                    self.process_frame(sws.scale(img), metadata)
                else:
                    print('Got unknown RemoteVideoPlayer message', msg, value)
            except Empty:
                break

    @error_guard
    def stop_listener(self, join=False):
        """Stops the client and also stops playing if we were playing.
        """
        self.stop(join=join)
        if self.listener_thread is None:
            return

        self.from_kivy_queue.put(('eof', None))
        if join:
            self.listener_thread.join()

        self.listener_thread = self.to_kivy_queue = self.from_kivy_queue = None
        self.client_active = False

    @error_guard
    def play(self):
        # we make sure client is connected and request metadata. If the player
        # is not playing on the server, we cannot ask it to play so we wait
        # until it sends us the metadata. If it was already playing, it sends
        # it again, otherwise, it'll send it to us when it starts playing
        self.start_listener()
        super(RemoteVideoPlayer, self).play()
        self.send_message_to_server('started_playing', None)

    def stop(self, *largs, join=False):
        if super(RemoteVideoPlayer, self).stop(join=join):
            self.send_message_to_server('stopped_playing', None)

    def play_thread_run(self):
        pass

    def stop_all(self, join=False):
        super(RemoteVideoPlayer, self).stop_all(join=join)
        self.stop_listener(join=join)


class ClientPlayerSettingsWidget(BoxLayout):
    """Settings widget for :class:`RemoteVideoPlayer`.
    """

    player: RemoteVideoPlayer = None
    """The player.
    """

    def __init__(self, player=None, **kwargs):
        if player is None:
            player = RemoteVideoPlayer()
        self.player = player
        super(ClientPlayerSettingsWidget, self).__init__(**kwargs)


Builder.load_file(join(dirname(__file__), 'client_player.kv'))



import time
import trio
from ..media_test_app import DemoTestApp


async def test_open_app(media_app: DemoTestApp, video_file):
    await media_app.wait_clock_frames(2)
    media_app.ffmpeg_player.play_filename = video_file
    media_app.ffmpeg_player.use_dshow = False
    media_app.player_name = 'ffmpeg'

    assert media_app.player == media_app.ffmpeg_player

    media_app._images_showed = 0
    ts = time.perf_counter()
    media_app.ffmpeg_player.play()
    while not media_app._images_showed:
        await trio.sleep(.1)
        if time.perf_counter() - ts >= 10:
            raise TimeoutError()

    media_app.ffmpeg_player.stop()



"""RTV24 based player
======================

This player can play RTV24 camera feeds using :mod:`pybarst`.
"""
from time import perf_counter as clock
import sys
import itertools
from os.path import splitext, join, exists, isdir, abspath, dirname, isfile

from ffpyplayer.pic import Image

from kivy.clock import Clock
from kivy.properties import (
    NumericProperty, StringProperty, BooleanProperty)
from kivy.uix.boxlayout import BoxLayout
from kivy.logger import Logger
from kivy.lang import Builder

from cpl_media.player import BasePlayer, VideoMetadata
from cpl_media import error_guard

try:
    import pybarst
    from pybarst.core.server import BarstServer
    from pybarst.rtv import RTVChannel
except ImportError as err:
    RTVChannel = BarstServer = None
    Logger.debug('cpl_media: Could not import pybarst: {}'.format(err))

__all__ = ('RTVPlayer', 'RTVSettingsWidget')


class RTVPlayer(BasePlayer):
    """Wrapper for RTV based player.
    """

    _config_props_ = ('remote_computer_name', 'pipe_name', 'port',
                          'video_fmt', 'pixel_fmt')

    video_fmts = {
        'full_NTSC': (640, 480), 'full_PAL': (768, 576),
        'CIF_NTSC': (320, 240), 'CIF_PAL': (384, 288),
        'QCIF_NTSC': (160, 120), 'QCIF_PAL': (192, 144)
    }
    """The video size formats supported by the player.
    """

    video_fmts_inverse = {v: k for k, v in video_fmts.items()}
    """Inverse of :attr:`video_fmts`.
    """

    image_fmts = {
        'rgb16': 'rgb565le', 'gray': 'gray', 'rgb15': 'rgb555le',
        'rgb24': 'rgb24', 'rgb32': 'rgba'}
    """The pixel formats from RTV -> ffmpeg, supported by the cameras.
    """

    remote_computer_name = StringProperty('')
    '''The name of the computer running Barst, if it's a remote computer.
    Otherwise it's the empty string.
    '''

    pipe_name = StringProperty('RTVPlayer')
    '''The internal name used to communicate with Barst. When running remotely,
    or if the server already is open, the name is used to discover Barst.
    '''

    port = NumericProperty(0)
    '''The RTV port (camera number) on the card to use.
    '''

    pixel_fmt = StringProperty('gray')
    '''The pixel format of the images being played.

    It can be one of the keys in :attr:`image_fmts`.
    '''

    video_fmt = StringProperty('full_NTSC')
    '''The video format of the video being played.

    It can be one of the keys in :attr:`video_fmts`.
    '''

    is_available = BooleanProperty(BarstServer is not None)
    """Whether pybarst is available to play."""

    barst_server = None
    """The :class:`pybarst.core.server.BarstServer` instance.
    """

    def __init__(self, **kwargs):
        super(RTVPlayer, self).__init__(**kwargs)

        self.fbind('remote_computer_name', self._update_summary)
        self.fbind('pipe_name', self._update_summary)
        self.fbind('port', self._update_summary)
        self._update_summary()

        self.fbind('video_fmt', self._update_metadata)
        self.fbind('pixel_fmt', self._update_metadata)
        self._update_metadata()

    def _update_summary(self, *largs):
        local = not self.remote_computer_name
        name = self.remote_computer_name if not local else '.'
        pipe_name = self.pipe_name

        self.player_summery = r'RTV "\\{}\pipe\{}:{}"'.format(
            name, pipe_name, self.port)

    def _update_metadata(self, *largs):
        w, h = self.video_fmts[self.video_fmt]
        pix_fmt = self.image_fmts[self.pixel_fmt]
        self.metadata_play = self.metadata_play_used = VideoMetadata(
            pix_fmt, w, h, 29.97)

    def play_thread_run(self):
        chan = None
        try:
            process_frame = self.process_frame
            paths = list(pybarst.dep_bins)
            if hasattr(sys, '_MEIPASS'):
                paths.append(sys._MEIPASS)

            barst_bin = None
            for p, f in itertools.product(paths, ('Barst64.exe', 'Barst.exe')):
                fname = join(abspath(p), f)
                if isfile(fname):
                    barst_bin = fname
                    break

            local = not self.remote_computer_name
            name = self.remote_computer_name if not local else '.'
            pipe_name = self.pipe_name
            full_name = r'\\{}\pipe\{}'.format(name, pipe_name)

            img_fmt = self.pixel_fmt
            ffmpeg_pix_fmt = self.image_fmts[img_fmt]
            w, h = self.video_fmts[self.video_fmt]
            video_fmt = self.video_fmt
            port = self.port

            if self.barst_server is None:
                self.barst_server = BarstServer(
                    barst_path=barst_bin, pipe_name=full_name)

            server = self.barst_server
            server.open_server()

            chan = RTVChannel(
                chan=port, server=server, video_fmt=video_fmt,
                frame_fmt=img_fmt, luma_filt=img_fmt == 'gray', lossless=True)

            chan.open_channel()
            try:
                chan.close_channel_server()
            except Exception:
                pass
            chan.open_channel()
            chan.set_state(True)

            started = False
            # use_rt = self.use_real_time
            count = 0
            ivl_start = 0

            while self.play_state != 'stopping':
                ts, buf = chan.read()
                if not started:
                    ivl_start = clock()
                    self.setattr_in_kivy_thread('ts_play', ivl_start)
                    Clock.schedule_once(self.complete_start)
                    started = True

                ivl_end = clock()
                if ivl_end - ivl_start >= 1.:
                    real_rate = count / (ivl_end - ivl_start)
                    self.setattr_in_kivy_thread('real_rate', real_rate)
                    count = 0
                    ivl_start = ivl_end

                count += 1
                self.increment_in_kivy_thread('frames_played')

                img = Image(
                    plane_buffers=[buf], pix_fmt=ffmpeg_pix_fmt, size=(w, h))
                process_frame(img, {'t': ts})
        except Exception as e:
            self.exception(e)
        finally:
            try:
                if chan is not None:
                    chan.close_channel_server()
            finally:
                Clock.schedule_once(self.complete_stop)

    @error_guard
    def stop_all(self, join=False):
        super(RTVPlayer, self).stop_all(join=join)

        barst_server = self.barst_server
        if barst_server is not None:
            barst_server.close_server()
            self.barst_server = None


class RTVSettingsWidget(BoxLayout):
    """Settings widget for :class:`RTVPlayer`.
    """

    player: RTVPlayer = None
    """The player.
    """

    def __init__(self, player=None, **kwargs):
        if player is None:
            player = RTVPlayer()
        self.player = player
        super(RTVSettingsWidget, self).__init__(**kwargs)


Builder.load_file(join(dirname(__file__), 'rtv_player.kv'))



"""cpl_media
=================

A library providing kivy support for playing and recording from
various cameras.
"""
from functools import wraps
import traceback
import sys
import pathlib
import os

__all__ = ('error_guard', 'error_callback', 'get_pyinstaller_datas')

__version__ = '0.1.3'


def _error_callback(e, exc_info=None, threaded=False):
    pass


error_callback = _error_callback
"""When set, care must be taken to handle errors from secondary threads.

It's signature is ``error_callback(e, exc_info=None, threaded=False)``.
"""


def error_guard(error_func):
    """A decorator which wraps the function in `try...except` and calls
    :func:`error_callback` if a exception is raised.

    E.g.::

        @error_guard
        def do_something():
            do_something_interesting
    """
    @wraps(error_func)
    def safe_func(*largs, **kwargs):
        try:
            return error_func(*largs, **kwargs)
        except Exception as e:
            exc_info = sys.exc_info()
            stack = traceback.extract_stack()
            tb = traceback.extract_tb(exc_info[2])
            full_tb = stack[:-1] + tb
            exc_line = traceback.format_exception_only(*exc_info[:2])

            err = 'Traceback (most recent call last):'
            err += "".join(traceback.format_list(full_tb))
            err += "".join(exc_line)
            error_callback(e, exc_info=err, threaded=True)

    return safe_func


def get_pyinstaller_datas():
    """Returns the ``datas`` list required by PyInstaller to be able to package
    :mod:`cpl_media` in a application.

    """
    root = pathlib.Path(os.path.dirname(sys.modules[__name__].__file__))
    datas = []
    for pat in ('**/*.kv', '*.kv'):
        for f in root.glob(pat):
            datas.append((str(f), str(f.relative_to(root.parent).parent)))

    return datas



"""Player
===========

Provides the base class for video players.
"""

import logging
from threading import Thread
from collections import namedtuple

from ffpyplayer.pic import get_image_size

from kivy.clock import Clock
from kivy.properties import (
    NumericProperty, ObjectProperty, StringProperty, BooleanProperty)
from kivy.event import EventDispatcher

from cpl_media import error_guard
from .common import KivyMediaBase

__all__ = ('BasePlayer', 'VideoMetadata')

VideoMetadata = namedtuple('VideoMetadata', ['fmt', 'w', 'h', 'rate'])
"""Namedtuple type describing a video stream.
"""


class BasePlayer(EventDispatcher, KivyMediaBase):
    """Base class for every player.
    """

    _config_props_ = ('metadata_play', 'metadata_play_used')

    display_frame = None
    """Called from kivy thread to display the frame whenever a new image
    arrives. This is called once per kivy frame, so if multiple images arrive
    during a kivy frame, it is called only for the last one.

    This callback takes two arguments: ``(image, metadata)``, where ``image``
    is the :class:`ffpyplayer.pic.Image`, and ``metadata`` is a dict with
    metadata.
    """

    display_trigger = None
    """Clock trigger to call :meth:`display_frame` in the kivy thread.
    """

    frame_callbacks = []
    """A list of callbacks that are called from the internal thread whenever
    a new image is available.

    All the callbacks are called with a single tuple argument
    ``(image, metadata)``, where ``image`` is the
    :class:`ffpyplayer.pic.Image`, and ``metadata`` is a dict with metadata.

    It always contains at least the key ``'t'`` indicating the timestamp of the
    image, but it may also contains other metadata keys specific to the player
    such as ``'count'`` for the frame number, when sent by the camera.
    """

    play_thread = None
    """The thread that plays the camera.
    """

    can_play = BooleanProperty(True)
    """Whether the video source can play now.
    """

    play_state = StringProperty('none')
    '''The current state of the state machine of the player.

    Can be one of none, starting, playing, stopping.
    '''

    last_image = None
    """The last :class:`ffpyplayer.pic.Image` received by the camera.
    """

    last_image_metadata = {'t': 0}
    """The metadata of the last image received by the camera.
    """

    use_real_time = False
    """Whether the video should use the current real time when we got the
    image, e.g. when the camera provided timestamp is not reliable.
    """

    metadata_play = ObjectProperty(None)
    '''(internal) Describes the video metadata of the video player. This is
    the requested format, or best guess of the metadata.

    Read only.
    '''

    metadata_play_used = ObjectProperty(None)
    '''(internal) Describes the video metadata of the video player that is
    actually used by the player. This must be set before recorders may allow
    recording the player.

    Depending on the metadata needed by the recorder, it may refuse to
    record until the needed metadata is given.

    Read only.
    '''

    real_rate = NumericProperty(0)
    """The estimated real fps of the video source being played.
    """

    frames_played = NumericProperty(0)
    """The number of frames that have been played since :meth:`play`
    was called.
    """

    ts_play = NumericProperty(0)
    """The time when the camera started playing.
    """

    player_summery = StringProperty('')
    """Textual summary of the camera type and config options.
    """

    data_rate = NumericProperty(0)
    """The estimated rate in B/s at which the camera is playing.
    """

    def __init__(self, **kwargs):
        self.frame_callbacks = []
        self.metadata_play = VideoMetadata(
            *kwargs.pop('metadata_play', ('', 0, 0, 0)))
        self.metadata_play_used = VideoMetadata(
            *kwargs.pop('metadata_play_used', ('', 0, 0, 0)))

        super(BasePlayer, self).__init__(**kwargs)
        self.display_trigger = Clock.create_trigger(self._display_frame, 0)

        self.fbind('metadata_play', self._update_data_rate)
        self.fbind('metadata_play_used', self._update_data_rate)
        self._update_data_rate()

    def _update_data_rate(self, *largs):
        fmt = self.metadata_play_used.fmt or self.metadata_play.fmt
        w = self.metadata_play_used.w or self.metadata_play.w
        h = self.metadata_play_used.h or self.metadata_play.h
        rate = self.metadata_play_used.rate or self.metadata_play.rate or 30
        if not fmt or not w or not h:
            self.data_rate = 0
        else:
            self.data_rate = sum(get_image_size(fmt, w, h)) * rate

    def _display_frame(self, *largs):
        if self.display_frame is not None:
            self.display_frame(self.last_image, self.last_image_metadata)

    def process_frame(self, frame, metadata):
        """Called from internal thread to process a new image frame received.

        :param frame: The :class:`ffpyplayer.pic.Image`.
        :param metadata: The metadata of the image. See
            :attr:`frame_callbacks`.
        """
        self.last_image = frame
        self.last_image_metadata = metadata
        for callback in self.frame_callbacks:
            callback((frame, metadata))
        self.display_trigger()

    def get_config_property(self, name):
        """(internal) used by the config system to get the special config data
        of the player.
        """
        if name == 'metadata_play':
            return tuple(self.metadata_play)
        if name == 'metadata_play_used':
            return tuple(self.metadata_play_used)
        return getattr(self, name)

    def apply_config_property(self, name, value):
        """(internal) used by the config system to set the special config data
        of the player.
        """
        if name == 'metadata_play':
            self.metadata_play = VideoMetadata(*value)
        elif name == 'metadata_play_used':
            self.metadata_play_used = VideoMetadata(*value)
        else:
            setattr(self, name, value)

    @error_guard
    def play(self):
        """Starts playing the video source and sets the :attr:`play_state` to
        `starting`.

        May be called from main kivy thread only.

        May only be called when :attr:`play_state` is `none`, otherwise an
        exception is raised.

        Players need to eventually call :meth:`complete_start` to finish
        starting playing.
        """
        if self.play_state != 'none' or not self.can_play:
            raise TypeError(
                'Asked to play while {} or disabled'.format(self.play_state))

        self.play_state = 'starting'
        self.ts_play = self.real_rate = 0.
        self.frames_played = 0
        thread = self.play_thread = Thread(
            target=self.play_thread_run, name='Play thread')
        thread.start()

    @error_guard
    def stop(self, *largs, join=False):
        """Stops playing the video source, if it is playing and sets the
        :attr:`play_state` to `stopping`.

        Players need to eventually call :meth:`complete_stop` to finish
        stopping playing.

        :param join: whether to block the thread until the internal play thread
            has exited.
        :return: Whether we stopped playing (True) or were already
            stop(ping/ed) playing.
        """
        if self.play_state == 'none':
            assert self.play_thread is None
            return False

        assert self.play_thread is not None

        if self.play_state == 'stopping':
            if join:
                self.play_thread.join()
            return False

        self.play_state = 'stopping'
        if join:
            self.play_thread.join()
        return True

    def update_metadata(self, fmt=None, w=None, h=None, rate=None):
        """Sets :attr:`metadata_play_used` based on :attr:`metadata_play` and
        any values provided to the function, if not None.
        """
        ifmt, iw, ih, irate = self.metadata_play
        if fmt is not None:
            ifmt = fmt
        if w is not None:
            iw = w
        if h is not None:
            ih = h
        if rate is not None:
            irate = rate

        self.metadata_play_used = VideoMetadata(ifmt, iw, ih, irate)

    def complete_start(self, *largs):
        """After :meth:`play`, this is called to set the player into `playing`
        :attr:`play_state`.
        """
        # when this is called, there may be a subsequent call scheduled
        # to stop, from the internal thread, but the internal thread never
        # requests first to stop and then to be in playing state
        assert self.play_state != 'none'
        # only internal thread sets to playing, and only once
        assert self.play_state != 'playing'
        if self.play_state == 'starting':  # not stopping
            self.play_state = 'playing'

    def complete_stop(self, *largs):
        """After :meth:`stop`, this is called to set the player into `none`
        :attr:`play_state`.
        """
        assert self.play_state != 'none'

        self.play_thread = None
        self.play_state = 'none'

    def play_thread_run(self):
        """The method that runs in the internal play thread.
        """
        raise NotImplementedError

    def stop_all(self, join=False):
        logging.info('cpl_media: Stopping all "{}"'.format(self))
        super(BasePlayer, self).stop_all(join=join)
        self.stop(join=join)



"""Demo App
===========

A demo app showing the configuration and usage of the players and recorders.

"""
from os.path import join, dirname
import math

from kivy.uix.boxlayout import BoxLayout
from kivy.properties import ObjectProperty, StringProperty, NumericProperty
from kivy.lang import Builder

from base_kivy_app.app import BaseKivyApp, run_app as run_base_app,\
    report_exception_in_app
from base_kivy_app.graphics import BufferImage
from cpl_media.ptgray import PTGrayPlayer, PTGraySettingsWidget
from cpl_media.ffmpeg import FFmpegPlayer, FFmpegSettingsWidget
from cpl_media.thorcam import ThorCamPlayer, ThorCamSettingsWidget
from cpl_media.remote.client import RemoteVideoPlayer, \
    ClientPlayerSettingsWidget
from cpl_media.rtv import RTVPlayer, RTVSettingsWidget
from cpl_media.player import BasePlayer

from cpl_media.recorder import ImageFileRecorder, VideoRecorder, \
    ImageFileRecordSettingsWidget, VideoRecordSettingsWidget
from cpl_media.remote.server import RemoteVideoRecorder, \
    RemoteRecordSettingsWidget
from cpl_media.recorder import BaseRecorder
import cpl_media


__all__ = ('DemoApp', 'run_app')

Builder.load_file(join(dirname(__file__), 'demo_app.kv'))


class RootAppWidget(BoxLayout):
    pass


class RecorderMetadata(BoxLayout):

    recorder = ObjectProperty(None, rebind=True)


class DemoApp(BaseKivyApp):

    _config_props_ = ('player_name', 'display_rotation')

    _config_children_ = {
        'ffmpeg': 'ffmpeg_player', 'ptgray': 'ptgray_player',
        'thor': 'thor_player', 'network_client': 'client_player',
        'rtv': 'rtv_player', 'image_file_recorder': 'image_file_recorder',
        'video_recorder': 'video_recorder',
        'network_server': 'server_recorder',
    }

    ffmpeg_player: FFmpegPlayer = None

    ffmpeg_settings = None

    ptgray_player: PTGrayPlayer = None

    ptgray_settings = None

    thor_player: ThorCamPlayer = None

    thor_settings = None

    client_player: RemoteVideoPlayer = None

    client_player_settings = None

    rtv_player: RTVPlayer = None

    rtv_settings = None

    player: BasePlayer = ObjectProperty(None, rebind=True)

    player_name = StringProperty('ffmpeg')

    image_display: BufferImage = None

    image_file_recorder: ImageFileRecorder = None

    image_file_recorder_settings = None

    video_recorder: VideoRecorder = None

    video_recorder_settings = None

    server_recorder: RemoteVideoRecorder = None

    server_recorder_settings = None

    display_rotation = NumericProperty(0)

    def build(self):
        self.ffmpeg_settings = FFmpegSettingsWidget()
        self.ffmpeg_player = self.ffmpeg_settings.player
        self.ptgray_settings = PTGraySettingsWidget()
        self.ptgray_player = self.ptgray_settings.player
        self.thor_settings = ThorCamSettingsWidget()
        self.thor_player = self.thor_settings.player
        self.client_player_settings = ClientPlayerSettingsWidget()
        self.client_player = self.client_player_settings.player
        self.rtv_settings = RTVSettingsWidget()
        self.rtv_player = self.rtv_settings.player

        self.image_file_recorder_settings = ImageFileRecordSettingsWidget()
        self.image_file_recorder = self.image_file_recorder_settings.recorder
        self.video_recorder_settings = VideoRecordSettingsWidget()
        self.video_recorder = self.video_recorder_settings.recorder
        self.server_recorder_settings = RemoteRecordSettingsWidget()
        self.server_recorder = self.server_recorder_settings.recorder

        self.load_app_settings_from_file()
        self.apply_app_settings()

        self.player = getattr(self, '{}_player'.format(self.player_name))

        self.ffmpeg_player.display_frame = self._display_frame
        self.ptgray_player.display_frame = self._display_frame
        self.thor_player.display_frame = self._display_frame
        self.client_player.display_frame = self._display_frame

        return super(DemoApp, self).build(root=RootAppWidget())

    def _display_frame(self, image, metadata):
        if self.image_display is not None:
            self.image_display.update_img(image)

    def stop_recording(self):
        self.server_recorder.stop()
        self.video_recorder.stop()
        self.image_file_recorder.stop()

    def clean_up(self):
        super(DemoApp, self).clean_up()
        for player in (
                self.ffmpeg_player, self.thor_player, self.client_player,
                self.image_file_recorder, self.video_recorder,
                self.server_recorder, self.rtv_player, self.ptgray_player):
            if player is not None:
                player.stop_all(join=True)
        self.dump_app_settings_to_file()


def run_app():
    """The function that starts the GUI and the entry point for
    the main script.
    """
    cpl_media.error_callback = report_exception_in_app
    return run_base_app(DemoApp)



"""Recorder
===========

Provides the base class for video recorders.
"""
from threading import Thread
from fractions import Fraction
from time import perf_counter as clock
from queue import Queue
from os.path import expanduser, join, exists, isdir, abspath, dirname

from ffpyplayer.pic import get_image_size, Image, SWScale
from ffpyplayer.tools import get_supported_pixfmts, get_format_codec
from ffpyplayer.writer import MediaWriter

from kivy.clock import Clock
from kivy.properties import (
    NumericProperty, ObjectProperty, StringProperty, BooleanProperty)
from kivy.event import EventDispatcher
from kivy.uix.boxlayout import BoxLayout
from kivy.lang import Builder

from .player import VideoMetadata, BasePlayer
from cpl_media import error_guard
from .common import KivyMediaBase

__all__ = ('BaseRecorder', 'ImageFileRecorder', 'VideoRecorder',
           'ImageFileRecordSettingsWidget', 'VideoRecordSettingsWidget')


class BaseRecorder(EventDispatcher, KivyMediaBase):
    """Records images from :class:cpl_media.player.BasePlayer` to a recorder.
    """

    _config_props_ = ('metadata_record', )

    player: BasePlayer = None
    """The :class:cpl_media.player.BasePlayer` this is being recorded from.
    """

    record_thread = None
    """The internal thread that records the frames from the player.
    """

    record_state = StringProperty('none')
    '''The current state of the state machine of the recorder.

    Can be one of none, starting, recording, stopping.

    State management:

    All state changes happen in the kivy main thread. It starts in none state.
    Requesting to record takes us out of none - we cannot get back to none
    until the internal thread requests it and we then get back to none.

    I.e. once out of none, the only way we could be back to none state is if
    all the internal thread requests has been processed. And only the internal
    thread is allowed to request we be back to none.
    '''

    image_queue = None
    """The queue used to communicate with the internal recording thread.
    """

    can_record = BooleanProperty(True)
    """Whether the recorder source can record now.
    """

    metadata_player = ObjectProperty(None)
    '''Describes the video metadata of the video player when we started
    recording.
    '''

    metadata_record = ObjectProperty(None)
    '''(internal) Describes the video metadata of the recorder. This is
    the requested format, or best guess of the metadata.

    Read only.
    '''

    metadata_record_used = ObjectProperty(None)
    '''(internal) Describes the video metadata of the recorder that is
    actually used by the recorder.

    Read only.
    '''

    frames_recorded = NumericProperty(0)
    """The number of frames recorded so far since :meth:`record`.
    """

    frames_skipped = NumericProperty(0)
    """The number of frames skipped and not recorded so far since
    :meth:`record`.
    """

    size_recorded = NumericProperty(0)
    """The estimated size of the data recorded so far since :meth:`record`.
    """

    ts_record = NumericProperty(0)
    """The time when the camera started recording.
    """

    data_rate = NumericProperty(0)
    """The estimated rate in B/s at which we're recording.
    """

    recorder_summery = StringProperty('')
    """Textual summary of the recorder type and config options.
    """

    elapsed_record_time = NumericProperty(0)
    """Number of seconds we have been recording since :meth:`record`.

    Automatically computed and updated a few times a second.
    """

    _elapsed_record_trigger = None

    def __init__(self, **kwargs):
        self.metadata_record_used = VideoMetadata('', 0, 0, 0)
        self.metadata_player = VideoMetadata(
            *kwargs.pop('metadata_player', ('', 0, 0, 0)))
        self.metadata_record = VideoMetadata(
            *kwargs.pop('metadata_record', ('', 0, 0, 0)))
        super(BaseRecorder, self).__init__(**kwargs)

        self._elapsed_record_trigger = Clock.create_trigger(
            self._update_elapsed_record, .2, True)

        self.fbind('metadata_player', self._update_data_rate)
        self.fbind('metadata_record', self._update_data_rate)
        self.fbind('metadata_record_used', self._update_data_rate)
        self._update_data_rate()

    def _update_elapsed_record(self, *largs):
        if self.ts_record:
            self.elapsed_record_time = clock() - self.ts_record

    def _update_data_rate(self, *largs):
        fmt = self.metadata_record_used.fmt
        w = self.metadata_record_used.w
        h = self.metadata_record_used.h
        rate = self.metadata_record_used.rate

        fmt = fmt or self.metadata_record.fmt or self.metadata_player.fmt
        w = w or self.metadata_record.w or self.metadata_player.w
        h = h or self.metadata_record.h or self.metadata_player.h
        rate = rate or self.metadata_record.rate or self.metadata_player.rate

        if not fmt or not w or not h:
            self.data_rate = 0
        else:
            rate = rate or 30
            self.data_rate = sum(get_image_size(fmt, w, h)) * rate

    def get_config_property(self, name):
        """(internal) used by the config system to get the special config data
        of the recorder.
        """
        if name == 'metadata_record':
            return tuple(self.metadata_record)
        return getattr(self, name)

    def apply_config_property(self, name, value):
        """(internal) used by the config system to set the special config data
        of the recorder.
        """
        if name == 'metadata_record':
            self.metadata_record = VideoMetadata(*value)
        else:
            setattr(self, name, value)

    @staticmethod
    def save_image(fname, img, codec='bmp', pix_fmt='', lib_opts={}):
        """Saves the given image to disk in the format requested.

        :param fname: The filename where to save the image.
        :param img: The :class:`ffpyplayer.pic.Image` to save.
        :param codec: The codec to pass to
            :class:`ffpyplayer.writer.MediaWriter` that determines the image
            type. Defaults to 'bmp'.
        :param pix_fmt: The pixel format into which to convert the image before
            saving. If empty, the original pixel format is used. If the
            codec doesn't support the image format, we first convert it to the
            closest supported format.
        :param lib_opts: Any additional `lib_opts` options to pass to
            :class:`ffpyplayer.writer.MediaWriter`.
        :return: The estimated size of the image on disk.
        """
        fmt = img.get_pixel_format()
        w, h = img.get_size()

        if not codec:
            codec = get_format_codec(fname)
            ofmt = get_supported_pixfmts(codec, fmt)[0]
        else:
            ofmt = get_supported_pixfmts(codec, pix_fmt or fmt)[0]
        if ofmt != fmt:
            sws = SWScale(w, h, fmt, ofmt=ofmt)
            img = sws.scale(img)
            fmt = ofmt

        out_opts = {'pix_fmt_in': fmt, 'width_in': w, 'height_in': h,
                    'frame_rate': (30, 1), 'codec': codec}
        writer = MediaWriter(fname, [out_opts], lib_opts=lib_opts)
        size = writer.write_frame(img=img, pts=0, stream=0)
        writer.close()
        return size

    @error_guard
    def record(self, player: BasePlayer):
        """Starts recording from the provided player and sets the
        :attr:`record_state` to `starting`.

        May be called from main kivy thread only.

        May only be called when :attr:`record_state` is `none`, otherwise an
        exception is raised. Similarly, only when the player's play state is
        `playing`. The players `metadata_play_used`, must also have been set to
        the value it's using.

        Recorders need to eventually call :meth:`complete_start` to finish
        starting recording.
        """
        if self.record_state != 'none':
            raise TypeError(
                'Asked to record while {}'.format(self.record_state))

        if player.play_state != 'playing':
            raise TypeError(
                'Can only record from player once the player is playing')

        self.record_state = 'starting'
        self.player = player
        self.size_recorded = self.ts_record = 0
        self.frames_recorded = self.frames_skipped = 0
        self.metadata_player = player.metadata_play_used
        self.image_queue = Queue()
        self._start_recording()

    def _start_recording(self):
        thread = self.record_thread = Thread(
            target=self.record_thread_run, name='Record thread')
        thread.start()

    @error_guard
    def stop(self, *largs, join=False):
        """Stops recording from the player, if it is recording and sets the
        :attr:`record_state` to `stopping`.

        Recorders need to eventually call :meth:`complete_stop` to finish
        stopping recording.

        :param join: whether to block the thread until the internal record
            thread has exited.
        :return: Whether we stopped recording (True) or were already
            stop(ping/ed) recording.
        """
        if self.record_state == 'none':
            assert self.record_thread is None
            return False

        assert self.image_queue is not None
        assert self.record_thread is not None

        if self.record_state == 'stopping':
            if join:
                self.record_thread.join()
            return False

        self.image_queue.put('eof')
        self.record_state = 'stopping'
        self._elapsed_record_trigger.cancel()
        if join:
            self.record_thread.join()
        return True

    def complete_start(self, *largs):
        """After :meth:`record`, this is called to set the recorder into
        `recording` :attr:`record_state`.
        """
        # when this is called, there may be a subsequent call scheduled
        # to stop, from the internal thread, but the internal thread never
        # requests first to stop and then to be in recording state
        assert self.record_state != 'none'
        # only internal thread sets to recording, and only once
        assert self.record_state != 'recording'
        if self.record_state == 'starting':  # not stopping
            self.record_state = 'recording'
        self._elapsed_record_trigger()

    def complete_stop(self, *largs):
        """After :meth:`stop`, this is called to set the recorder into `none`
        :attr:`record_state`.
        """
        assert self.record_state != 'none'

        self.record_thread = None
        self.image_queue = None
        self.record_state = 'none'

    def record_thread_run(self, *largs):
        """The method that runs in the internal record thread.
        """
        raise NotImplementedError

    def stop_all(self, join=False):
        super(BaseRecorder, self).stop_all(join=join)
        self.stop(join=join)


class ImageFileRecorder(BaseRecorder):
    """Records images as files to disk.
    """

    _config_props_ = (
        'record_directory', 'record_prefix', 'compression', 'extension')

    record_directory = StringProperty(expanduser('~'))
    '''The directory into which videos should be saved.
    '''

    record_prefix = StringProperty('image_')
    '''The prefix to the filename of the images being saved.
    '''

    extension = StringProperty('tiff')
    """The extension of the images being saved.
    """

    compression = StringProperty('raw')
    """Whether to compress when :attr:`extension` is `tiff`. Can be one of
    ``'raw', 'lzw', 'zip'``.
    """

    def __init__(self, **kwargs):
        super(ImageFileRecorder, self).__init__(**kwargs)

        self.fbind('record_directory', self._update_summary)
        self.fbind('record_prefix', self._update_summary)
        self.fbind('extension', self._update_summary)
        self._update_summary()

    def _update_summary(self, *largs):
        self.recorder_summery = 'FFmpeg "{}*.{}"'.format(
            join(self.record_directory, self.record_prefix), self.extension)

    def send_image_to_recorder(self, image):
        """Sends the image to the recorder queue to save the image.

        :param image: A tuple of the image and metadata as provided to
            :attr:`cpl_media.player.BasePlayer.frame_callbacks`.
        """
        if self.image_queue is None:
            return

        self.image_queue.put(image)

    @error_guard
    def record(self, player: BasePlayer):
        if not player.metadata_play_used or not player.metadata_play_used.rate:
            raise TypeError(
                'Can only record from player once the fps is known')

        super(ImageFileRecorder, self).record(player=player)
        self.player.frame_callbacks.append(self.send_image_to_recorder)

    @error_guard
    def stop(self, *largs, join=False):
        if super(ImageFileRecorder, self).stop(join=join):
            self.player.frame_callbacks.remove(self.send_image_to_recorder)

    def _start_recording(self):
        self.record_directory = expanduser(self.record_directory)
        thread = self.record_thread = Thread(
            target=self.record_thread_run, name='Record image thread',
            args=(self.record_directory, self.record_prefix, self.compression,
                  self.extension))
        thread.start()

    def complete_stop(self, *largs):
        super(ImageFileRecorder, self).complete_stop()
        if self.send_image_to_recorder in self.player.frame_callbacks:
            self.player.frame_callbacks.remove(self.send_image_to_recorder)

    def record_thread_run(
            self, record_directory, record_prefix, compression, extension):
        queue = self.image_queue
        last_img = None

        while self.record_state != 'stopping':
            item = queue.get()
            if item == 'eof':
                break
            image, metadata = item

            try:
                if last_img is None:
                    self.setattr_in_kivy_thread('ts_record', clock())
                    self.setattr_in_kivy_thread(
                        'metadata_record_used', self.player.metadata_play_used)
                    Clock.schedule_once(self.complete_start)
                    last_img = image

                suffix = 't={}'.format(metadata['t'])
                if 'count' in metadata:
                    suffix += '_count={}'.format(metadata['count'])
                ext = '.' + extension
                filename = join(record_directory, record_prefix + suffix + ext)
                counter = 0
                while exists(filename):
                    counter += 1
                    filename = join(
                        record_directory,
                        record_prefix + suffix + '-{}'.format(counter) + ext)

                if extension == 'tiff':
                    lib_opts = {
                        'compression_algo': 'deflate' if compression == 'zip'
                        else compression}
                else:
                    lib_opts = {}

                size = self.save_image(
                    filename, image, codec=extension,
                    pix_fmt=image.get_pixel_format(), lib_opts=lib_opts)
                self.increment_in_kivy_thread('size_recorded', size)
                self.increment_in_kivy_thread('frames_recorded')
            except Exception as e:
                self.exception(e)
                self.increment_in_kivy_thread('frames_skipped')

        Clock.schedule_once(self.complete_stop)


class VideoRecorder(BaseRecorder):
    """Records images to a video file on disk.

    Cannot start recording until the player fps is known. Otherwise, an error
    is raised.
    """

    _config_props_ = (
        'record_directory', 'record_fname', 'record_fname_count',
        'estimate_record_rate')

    record_directory = StringProperty(expanduser('~'))
    '''The directory into which videos should be saved.
    '''

    record_fname = StringProperty('video{}.mkv')
    '''The filename to be used to record the next video.

    If ``{}`` is present in the filename, it'll be replaced with the value of
    :attr:`record_fname_count` which auto increments after every video, when
    used. We use Python's ``format`` machinery for this.
    '''

    record_fname_count = NumericProperty(0)
    '''A counter that auto increments by one after every recorded video.

    Used to give unique filenames for each video file. See
    :attr:`record_fname`.
    '''

    estimate_record_rate = BooleanProperty(False)
    """Whether to use :attr:`cpl_media.player.BasePlayer.real_rate` for the
    recorder frame rate, as opposed to the one initially provided by the
    player.
    """

    record_filename = StringProperty('')
    """The full filename of the video that will be saved as computed by the
    configuration options of the instance.
    """

    def __init__(self, **kwargs):
        super(VideoRecorder, self).__init__(**kwargs)

        self.fbind('record_directory', self._update_record_fname)
        self.fbind('record_fname', self._update_record_fname)
        self.fbind('record_fname_count', self._update_record_fname)
        self._update_record_fname()

        self.fbind('record_filename', self._update_summary)
        self._update_summary()

    def _update_summary(self, *largs):
        self.recorder_summery = 'FFmpeg "{}"'.format(self.record_filename)

    def compute_recording_opts(self, ifmt=None, iw=None, ih=None):
        """Computes the recording metadata to use, from the provided options
        and from the player's metadata.
        """
        play_used = self.metadata_player
        ifmt = ifmt or play_used.fmt
        iw = iw or play_used.w
        ih = ih or play_used.h
        irate = play_used.rate

        ifmt = ifmt or 'yuv420p'
        iw = iw or 640
        ih = ih or 480
        assert irate

        if self.record_fname.endswith('mkv') and (
                ifmt.startswith('rgb') or ifmt.startswith('bgr')):
            ifmt = get_supported_pixfmts('libx264', ifmt)

        ofmt, ow, oh, orate = self.metadata_record
        ofmt = ofmt or ifmt
        ow = ow or iw
        oh = oh or ih
        if self.estimate_record_rate:
            orate = orate or self.player.real_rate
        orate = orate or irate

        return (ifmt, iw, ih, irate), (ofmt, ow, oh, orate)

    def _set_metadata_record(self, fmt, w, h, rate):
        self.metadata_record = VideoMetadata(fmt, w, h, rate)

    def _update_record_fname(self, *largs):
        self.record_filename = join(
            self.record_directory,
            self.record_fname.format(self.record_fname_count))

    def send_image_to_recorder(self, image):
        """Sends the image to the recorder queue to save the image in the
        video.

        :param image: A tuple of the image and metadata as provided to
            :attr:`cpl_media.player.BasePlayer.frame_callbacks`.
        """
        if self.image_queue is None:
            return

        self.image_queue.put(image)

    @error_guard
    def record(self, player: BasePlayer):
        if not player.metadata_play_used or not player.metadata_play_used.rate:
            raise TypeError(
                'Can only record from player once the fps is known')

        super(VideoRecorder, self).record(player=player)
        self.player.frame_callbacks.append(self.send_image_to_recorder)

    @error_guard
    def stop(self, *largs, join=False):
        if super(VideoRecorder, self).stop(join=join):
            self.player.frame_callbacks.remove(self.send_image_to_recorder)

    def _start_recording(self):
        self.record_directory = expanduser(self.record_directory)
        thread = self.record_thread = Thread(
            target=self.record_thread_run, name='Record thread',
            args=(self.record_filename, ))
        thread.start()

    def complete_stop(self, *largs):
        super(VideoRecorder, self).complete_stop()
        self.record_fname_count += 1
        if self.send_image_to_recorder in self.player.frame_callbacks:
            self.player.frame_callbacks.remove(self.send_image_to_recorder)

    def record_thread_run(self, filename):
        queue = self.image_queue
        recorder = None
        t0 = None
        last_t = None

        while self.record_state != 'stopping':
            item = queue.get()
            if item == 'eof':
                break
            img, metadata = item

            if recorder is None:
                try:
                    self.setattr_in_kivy_thread('ts_record', clock())
                    t0 = metadata['t']
                    iw, ih = img.get_size()
                    ipix_fmt = img.get_pixel_format()

                    (ifmt, iw, ih, irate), (opix_fmt, ow, oh, orate) = \
                        self.compute_recording_opts(ipix_fmt, iw, ih)

                    self.setattr_in_kivy_thread(
                        'metadata_record_used',
                        VideoMetadata(opix_fmt, ow, oh, orate))

                    orate = Fraction(orate)
                    if orate >= 1.:
                        orate = Fraction(orate.denominator, orate.numerator)
                        orate = orate.limit_denominator(2 ** 30 - 1)
                        orate = (orate.denominator, orate.numerator)
                    else:
                        orate = orate.limit_denominator(2 ** 30 - 1)
                        orate = (orate.numerator, orate.denominator)

                    stream = {
                        'pix_fmt_in': ipix_fmt, 'pix_fmt_out': opix_fmt,
                        'width_in': iw, 'height_in': ih, 'width_out': ow,
                        'height_out': oh, 'codec': 'rawvideo',
                        'frame_rate': orate}

                    recorder = MediaWriter(filename, [stream])
                except Exception as e:
                    self.exception(e)
                    Clock.schedule_once(self.complete_stop)
                    return

                Clock.schedule_once(self.complete_start)

            try:
                self.setattr_in_kivy_thread(
                    'size_recorded',
                    recorder.write_frame(img, metadata['t'] - t0))
                self.increment_in_kivy_thread('frames_recorded')
            except Exception as e:
                self.exception(e)
                self.increment_in_kivy_thread('frames_skipped')

        if recorder is not None:
            try:
                recorder.close()
            except Exception as e:
                self.exception(e)

        Clock.schedule_once(self.complete_stop)


class ImageFileRecordSettingsWidget(BoxLayout):
    """Settings widget for :class:`ImageFileRecorder`.
    """

    recorder: ImageFileRecorder = None
    """The recorder.
    """

    def __init__(self, recorder=None, **kwargs):
        if recorder is None:
            recorder = ImageFileRecorder()
        self.recorder = recorder
        super(ImageFileRecordSettingsWidget, self).__init__(**kwargs)

    def set_filename(self, text_wid, paths):
        """Called by the GUI to set the directory.
        """
        if not paths:
            return

        self.recorder.record_directory = paths[0]
        text_wid.text = paths[0]


class VideoRecordSettingsWidget(BoxLayout):
    """Settings widget for :class:`VideoRecorder`.
    """

    recorder: VideoRecorder = None
    """The recorder.
    """

    def __init__(self, recorder=None, **kwargs):
        if recorder is None:
            recorder = VideoRecorder()
        self.recorder = recorder
        super(VideoRecordSettingsWidget, self).__init__(**kwargs)

    def set_filename(self, text_wid, paths):
        """Called by the GUI to set the directory.
        """
        if not paths:
            return

        self.recorder.record_directory = paths[0]
        text_wid.text = paths[0]


Builder.load_file(join(dirname(__file__), 'recorder.kv'))



"""FFmpeg based player
======================

:class:`FFmpegPlayer` can play USB cameras, video files, etc using
:mod:`ffpyplayer`.

"""
import re
import time
from collections import defaultdict
from functools import partial
from time import perf_counter as clock
from os.path import splitext, join, exists, isdir, abspath, dirname

from ffpyplayer.player import MediaPlayer
from ffpyplayer.pic import get_image_size
from ffpyplayer.tools import list_dshow_devices, set_log_callback

from kivy.clock import Clock
from kivy.logger import Logger
from kivy.properties import StringProperty, DictProperty, BooleanProperty, \
    NumericProperty
from kivy.uix.boxlayout import BoxLayout
from kivy.lang import Builder

from cpl_media.player import BasePlayer, VideoMetadata
from cpl_media import error_guard

__all__ = ('FFmpegPlayer', 'FFmpegSettingsWidget', 'LogFilter')

set_log_callback(logger=Logger, default_only=True)


def eat_first(f, val, *largs, **kwargs):
    f(*largs, **kwargs)


class FFmpegPlayer(BasePlayer):
    """Wrapper for :mod:`ffpyplayer` ffmpeg based player.
    """

    _config_props_ = (
        'play_filename', 'file_fmt', 'icodec',
        'dshow_true_filename', 'dshow_opt', 'use_dshow', 'dshow_rate',
        'dshow_filename')

    play_filename = StringProperty('')
    '''The filename of the media being played. Can be e.g. a filename etc.
    '''

    file_fmt = StringProperty('')
    '''The format used to play the video. Can be empty or a format e.g.
    ``mjpeg`` for webcams.
    '''

    icodec = StringProperty('')
    '''The codec used to open the video stream with if it needs to be
    specified for the camera.
    '''

    use_dshow = BooleanProperty(True)
    """Whether we use dshow - i.e. USB webcams, or normal media sources.
    """

    dshow_rate = NumericProperty(0)
    """The frame rate to request from the dshow camera.
    """

    dshow_filename = StringProperty('')
    """The name of the dshow camera to open.
    """

    dshow_true_filename = StringProperty('')
    '''The real and complete filename of the direct show (webcam) device.
    '''

    dshow_opt = StringProperty('')
    '''The camera options associated with :attr:`dshow_true_filename` when
    dshow is used.
    '''

    dshow_names = DictProperty({})
    """All the cameras that can be opened.
    """

    dshow_opts = DictProperty({})
    """The options supported by the cameras that can be opened.
    """

    dshow_opt_pat = re.compile(
        '([0-9]+)X([0-9]+) (.+), ([0-9.]+)(?: - ([0-9.]+))? fps')

    _config_dshow_filename = ''

    def __init__(self, **kw):
        super(FFmpegPlayer, self).__init__(**kw)

        self.fbind('play_filename', self._update_summary)
        self.fbind('dshow_filename', self._update_summary)
        self.fbind('file_fmt', self._update_summary)
        self.fbind('use_dshow', self._update_summary)
        self._update_summary()

    def _set_metadata_play(self, fmt, w, h):
        self.metadata_play = VideoMetadata(fmt, w, h, 0)

    def _update_summary(self, *largs):
        name = self.dshow_filename if self.use_dshow else self.play_filename
        self.player_summery = 'FFmpeg "{}"'.format(name)

    def apply_config_property(self, name, value):
        if name == 'dshow_filename':
            self._config_dshow_filename = value
        else:
            super(FFmpegPlayer, self).apply_config_property(name, value)

    def post_config_applied(self):
        """Handles settings as applied by the app config system so the
        properties are set to correct values.
        """
        # this must be set after everything so we can loop up its opts in dict
        dshow_filename = self._config_dshow_filename

        try:
            if self.dshow_opt:
                self._parse_dshow_opt(self.dshow_opt)
        except Exception:
            self.dshow_opt = ''

        if not self.dshow_true_filename or not dshow_filename or not \
                self.dshow_opt:
            dshow_filename = self.dshow_true_filename = ''
            self.dshow_opt = ''

        if dshow_filename:
            if dshow_filename not in self.dshow_names:
                self.dshow_names[dshow_filename] = \
                    self.dshow_true_filename

            if dshow_filename not in self.dshow_opts:
                self.dshow_opts[dshow_filename] = {}

            self.dshow_opts[dshow_filename][self.dshow_opt] = \
                self._parse_dshow_opt(self.dshow_opt)

        self.dshow_filename = dshow_filename
        super().post_config_applied()

    @error_guard
    def refresh_dshow(self):
        """Refreshes list of direct show cameras available.
        """
        counts = defaultdict(int)
        video, _, names = list_dshow_devices()
        video2 = {}
        names2 = {}

        # rename to have pretty unique names
        for true_name, name in names.items():
            if true_name not in video:
                continue

            count = counts[name]
            name2 = '{}-{}'.format(name, count) if count else name
            counts[name] = count + 1

            # filter and clean cam opts
            names2[name2] = true_name
            opts = video2[name2] = {}

            for fmt, _, (w, h), (rmin, rmax) in video[true_name]:
                if not fmt:
                    continue
                if rmin != rmax:
                    key = '{}X{} {}, {} - {} fps'.format(w, h, fmt, rmin, rmax)
                else:
                    key = '{}X{} {}, {} fps'.format(w, h, fmt, rmin)
                if key not in opts:
                    opts[key] = (fmt, (w, h), (rmin, rmax))

        self.dshow_opts = video2
        self.dshow_names = names2
        if self.dshow_filename not in names2:
            if not names2:
                self.dshow_filename = ''
                self.dshow_true_filename = ''
            else:
                self.dshow_filename = list(names2.keys())[0]
                self.dshow_true_filename = names2[self.dshow_filename]
        self.update_dshow_file()

    @error_guard
    def update_dshow_file(self):
        """Updates the dshow camera name and options in response to a
        re-configuration.
        """
        if not self.use_dshow or not self.dshow_filename:
            self.dshow_opt = ''
            self.dshow_true_filename = ''
            return

        assert self.dshow_filename in self.dshow_opts
        self.dshow_true_filename = self.dshow_names[self.dshow_filename]
        if self.dshow_opt in self.dshow_opts[self.dshow_filename]:
            return
        opts = list(self.dshow_opts[self.dshow_filename].keys())
        if not opts:
            self.dshow_opt = ''
        else:
            self.dshow_opt = opts[0]

    @error_guard
    def parse_dshow_opt(self, opt):
        """Parses the :attr:`dshow_opts` type string option into
        `(fmt, (w, h), (rmin, rmax))`
        """
        return self._parse_dshow_opt(opt)

    def _parse_dshow_opt(self, opt):
        m = re.match(self.dshow_opt_pat, opt)
        if m is None:
            raise ValueError('{} not a valid option'.format(opt))

        w, h, fmt, rmin, rmax = m.groups()
        if rmax is None:
            rmax = rmin

        w, h, rmin, rmax = int(w), int(h), float(rmin), float(rmax)
        return fmt, (w, h), (rmin, rmax)

    def get_opt_image_size(self, opt):
        fmt, (w, h), _ = self.parse_dshow_opt(opt)
        return w * h, sum(get_image_size(fmt, w, h))

    @error_guard
    def player_callback(self, mode, value):
        """Called internally by ffpyplayer when an error occurs internally.
        """
        if mode.endswith('error'):
            raise Exception(
                'FFmpeg Player: internal error "{}", "{}"'.format(mode, value))

    def play_thread_run(self):
        try:
            self._play_thread_run()
        except Exception as e:
            self.exception(e)
        finally:
            Clock.schedule_once(self.complete_stop)

    def _play_thread_run(self):
        process_frame = self.process_frame
        ff_opts = {'sync': 'video', 'an': True, 'sn': True, 'paused': True}

        ifmt, icodec = self.file_fmt, self.icodec
        use_dshow = self.use_dshow
        if ifmt:
            ff_opts['f'] = ifmt
        if use_dshow:
            ff_opts['f'] = 'dshow'
        if icodec:
            ff_opts['vcodec'] = icodec

        ipix_fmt, iw, ih, _ = self.metadata_play
        ff_opts['x'] = iw
        ff_opts['y'] = ih

        lib_opts = {}
        if use_dshow:
            rate = self.dshow_rate
            if self.dshow_opt:
                fmt, size, (rmin, rmax) = self.parse_dshow_opt(self.dshow_opt)
                lib_opts['pixel_format'] = fmt
                lib_opts['video_size'] = '{}x{}'.format(*size)
                if rate:
                    rate = min(max(rate, rmin), rmax)
                    lib_opts['framerate'] = '{}'.format(rate)
            elif rate:
                lib_opts['framerate'] = '{}'.format(rate)

        fname = self.play_filename
        if use_dshow:
            fname = 'video={}'.format(self.dshow_true_filename)

        ffplayer = MediaPlayer(
            fname, callback=self.player_callback, ff_opts=ff_opts,
            lib_opts=lib_opts)

        # wait for media to init pixel fmt
        src_fmt = ''
        s = clock()
        while self.play_state == 'starting' and clock() - s < 5.:
            src_fmt = ffplayer.get_metadata().get('src_pix_fmt')
            if src_fmt:
                break
            time.sleep(0.01)

        if not src_fmt:
            raise ValueError("Player failed, couldn't get pixel type")

        if ipix_fmt:
            src_fmt = ipix_fmt
        fmt = {'gray': 'gray', 'rgb24': 'rgb24', 'bgr24': 'rgb24',
               'rgba': 'rgba', 'bgra': 'rgba'}.get(src_fmt, 'yuv420p')
        ffplayer.set_output_pix_fmt(fmt)

        ffplayer.toggle_pause()
        Logger.info('FFmpeg Player: input, output formats are: {}, {}'.
                    format(src_fmt, fmt))

        # wait for first frame
        img = None
        s = clock()
        ivl_start = None
        while self.play_state == 'starting' and clock() - s < 5.:
            img, val = ffplayer.get_frame()
            if val == 'eof':
                raise ValueError("Player failed, reached eof")

            if img:
                ivl_start = clock()
                break
            time.sleep(0.01)

        rate = ffplayer.get_metadata().get('frame_rate')
        if rate == (0, 0) or not rate or not rate[1]:
            raise ValueError("Player failed, couldn't read frame rate")

        if not img:
            raise ValueError("Player failed, couldn't read frame")

        # ready to start
        rate = rate[0] / float(rate[1])
        w, h = img[0].get_size()
        fmt = img[0].get_pixel_format()
        use_rt = self.use_real_time

        Clock.schedule_once(
            partial(eat_first, self.update_metadata, rate=rate, w=w, h=h,
                    fmt=fmt), 0)
        Clock.schedule_once(self.complete_start)

        # started
        process_frame(img[0], {'t': ivl_start if use_rt else img[1]})

        min_sleep = 1 / (rate * 8.)
        self.setattr_in_kivy_thread('ts_play', ivl_start)
        self.setattr_in_kivy_thread('frames_played', 1)
        count = 1

        while self.play_state != 'stopping':
            img, val = ffplayer.get_frame()
            ivl_end = clock()

            if ivl_end - ivl_start >= 1.:
                real_rate = count / (ivl_end - ivl_start)
                self.setattr_in_kivy_thread('real_rate', real_rate)
                count = 0
                ivl_start = ivl_end

            if val == 'paused':
                raise ValueError("Player {} got {}".format(self, val))
            if val == 'eof':
                break

            if not img:
                time.sleep(min(val, min_sleep) if val else min_sleep)
                continue
            elif val:
                ts = clock()
                leftover = val
                while leftover > min_sleep and \
                        self.play_state != 'stopping':
                    time.sleep(min_sleep)
                    leftover = max(val - (clock() - ts), 0)

            count += 1
            self.increment_in_kivy_thread('frames_played')
            process_frame(img[0], {'t': ivl_end if use_rt else img[1]})


class FFmpegSettingsWidget(BoxLayout):
    """Settings widget for :class:`FFmpegPlayer`.
    """

    player: FFmpegPlayer = None
    """The player.
    """

    def __init__(self, player=None, **kwargs):
        if player is None:
            player = FFmpegPlayer()
        self.player = player
        super(FFmpegSettingsWidget, self).__init__(**kwargs)

    def set_filename(self, text_wid, paths):
        """Called by the GUI to set the filename.
        """
        if not paths:
            return

        self.player.play_filename = paths[0]
        text_wid.text = paths[0]


class LogFilter:
    """FFmpeg log filter that removes specific messages that are repeated.
    """

    _last_msg: str = ''

    _logger_func: dict = {}

    filtered_end: str = '[rtbufsize parameter])! frame dropped!'
    """Remove repeated messages that end with this string.
    """

    def __init__(self):
        self._logger_func = {
            'quiet': Logger.critical, 'panic': Logger.critical,
            'fatal': Logger.critical, 'error': Logger.error,
            'warning': Logger.warning, 'info': Logger.info,
            'verbose': Logger.debug, 'debug': Logger.debug,
            'trace': Logger.debug
        }

    def start_filter(self):
        """Start the filter.
        """
        set_log_callback(self._callback)

    def _callback(self, message: str, level: str):
        message = message.strip()
        end = self.filtered_end
        if not message or end and message.endswith(end) and \
                self._last_msg.endswith(end):
            return

        self._last_msg = message
        self._logger_func[level](f'ffpyplayer: {message}')


Builder.load_file(join(dirname(__file__), 'ffmpeg_player.kv'))



import os
import pytest
import trio
import time
import warnings
from .media_test_app import DemoTestApp

warnings.filterwarnings(
    "ignore",
    message="numpy.ufunc size changed, may indicate binary incompatibility. "
            "Expected 192 from C header, got 216 from PyObject"
)

os.environ['KIVY_USE_DEFAULTCONFIG'] = '1'


@pytest.fixture()
async def media_app(request, nursery, tmp_path):
    ts0 = time.perf_counter()
    from kivy.core.window import Window
    from kivy.context import Context
    from kivy.clock import ClockBase
    from kivy.base import stopTouchApp
    from kivy.logger import LoggerHistory

    context = Context(init=False)
    context['Clock'] = ClockBase(async_lib='trio')
    context.push()

    Window.create_window()
    Window.register()
    Window.initialized = True
    Window.canvas.clear()

    from more_kivy_app.app import report_exception_in_app
    import cpl_media
    cpl_media.error_callback = report_exception_in_app

    app = DemoTestApp(
        yaml_config_path=str(tmp_path / 'config.yaml'),
        ini_file=str(tmp_path / 'config.ini'))

    try:
        app.set_async_lib('trio')
        nursery.start_soon(app.async_run)

        ts = time.perf_counter()
        while not app.app_has_started:
            await trio.sleep(.1)
            if time.perf_counter() - ts >= 10:
                raise TimeoutError()

        await app.wait_clock_frames(5)

        ts1 = time.perf_counter()
        yield app
        ts2 = time.perf_counter()

        stopTouchApp()

        ts = time.perf_counter()
        while not app.app_has_stopped:
            await trio.sleep(.1)
            if time.perf_counter() - ts >= 10:
                raise TimeoutError()
    finally:
        stopTouchApp()
        app.clean_up()
        for child in Window.children[:]:
            Window.remove_widget(child)

        context.pop()
        del context
        LoggerHistory.clear_history()

    ts3 = time.perf_counter()
    print(ts1 - ts0, ts2 - ts1, ts3 - ts2)


@pytest.fixture(scope='session')
def video_file(tmp_path_factory):
    from ffpyplayer.writer import MediaWriter
    from ffpyplayer.pic import Image
    fname = str(tmp_path_factory.mktemp('data') / 'test_video.avi')

    w, h = 64, 64
    size = w * h
    out_opts = {
        'pix_fmt_in': 'gray', 'width_in': w, 'height_in': h,
        'codec': 'rawvideo', 'frame_rate': (2997, 100)}

    buf = bytearray([int(x * 255 / size) for x in range(size)])
    img = Image(plane_buffers=[buf], pix_fmt='gray', size=(w, h))

    writer = MediaWriter(fname, [out_opts])
    for i in range(20):
        writer.write_frame(img=img, pts=i / 29.97, stream=0)
    writer.close()

    return fname



from setuptools import setup, find_packages
from io import open
from os import path

from cpl_media import __version__

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

URL = 'https://github.com/matham/cpl_media'

setup(
    name='cpl_media',
    version=__version__,
    author='Matthew Einhorn',
    author_email='moiein2000@gmail.com',
    license='MIT',
    description=(
        'Kivy support for playing and recording various cameras.'),
    long_description=long_description,
    long_description_content_type='text/markdown',
    url=URL,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    packages=find_packages(),
    install_requires=[
        'base_kivy_app~=0.1.1', 'ffpyplayer', 'kivy', 'tree-config'],
    extras_require={
        'dev': ['pytest>=3.6', 'pytest-cov', 'flake8', 'sphinx-rtd-theme',
                'coveralls', 'trio', 'pytest-trio'],
    },
    package_data={'cpl_media': ['*.kv', '**/*.kv']},
    project_urls={
        'Bug Reports': URL + '/issues',
        'Source': URL,
    },
    entry_points={
        'console_scripts': ['cpl_media=cpl_media.tests.app.demo_app:run_app']},
)



"""Base classes
=========================

Provides common base classes and tools.

"""

from queue import Queue, Empty
import sys
import traceback
from more_kivy_app.config import Configurable

from kivy.clock import Clock

from cpl_media import error_guard
import cpl_media

__all__ = ('KivyMediaBase', )


class KivyMediaBase(Configurable):
    """A base classes for all the players and recorders.

    It provides methods for the kivy and internal threads to interact safely.
    Specifically, for the internal threads to schedule code to be executed in
    the kivy thread.
    """

    trigger_run_in_kivy = None
    """A Kivy clock trigger that will cause
    :meth:`process_queue_in_kivy_thread` to be called on the next frame in the
    kivy thread.
    """

    kivy_thread_queue = None
    """The queue that the kivy thread will read from and process messages.
    """

    def __init__(self, **kwargs):
        super(KivyMediaBase, self).__init__(**kwargs)
        self.kivy_thread_queue = Queue()
        self.trigger_run_in_kivy = Clock.create_trigger(
            self.process_queue_in_kivy_thread)

    @error_guard
    def process_queue_in_kivy_thread(self, *largs):
        """Method that is called in the kivy thread when
        :attr:`trigger_run_in_kivy` is triggered. It reads messages from the
        thread.
        """
        while self.kivy_thread_queue is not None:
            try:
                msg, value = self.kivy_thread_queue.get(block=False)

                if msg == 'setattr':
                    prop, val = value
                    setattr(self, prop, val)
                elif msg == 'increment':
                    prop, val = value
                    setattr(self, prop, getattr(self, prop) + val)
                else:
                    print('Got unknown KivyMediaBase message', msg, value)
            except Empty:
                break

    def setattr_in_kivy_thread(self, prop, value):
        """Schedules kivy to set the property to the specified value in the
        kivy thread.

        :param prop: The instance property name to set.
        :param value: The value the property will be set to.
        """
        self.kivy_thread_queue.put(('setattr', (prop, value)))
        self.trigger_run_in_kivy()

    def increment_in_kivy_thread(self, prop, value=1):
        """Schedules kivy to increment the property by the specified value in
        the kivy thread.

        :param prop: The instance property name to increment.
        :param value: The value by which it will be incremented.
        """
        self.kivy_thread_queue.put(('increment', (prop, value)))
        self.trigger_run_in_kivy()

    def stop_all(self, join=False):
        """Causes all internal threads to stop and exit.

        :param join: Whether to wait and block the calling thread until the
            internal threads exit.
        """
        pass

    def exception(self, e):
        """Called from under an exception, to report the exception to
        :func:`cpl_media.error_callback`.

        :param e: The exception instance.
        """
        cpl_media.error_callback(
            e, exc_info=''.join(traceback.format_exception(*sys.exc_info())),
            threaded=True)
